# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import ast
import _ast
import os
import sys

from . import package_root, task_keyword_args
from ._import import _import_from


if sys.version_info < (3,):
    byte_cls = str
else:
    byte_cls = bytes


def _list_tasks():
    """
    Fetches a list of all valid tasks that may be run, and the args they
    accept. Does not actually import the task module to prevent errors if a
    user does not have the dependencies installed for every task.

    :return:
        A list of 2-element tuples:
         0: a unicode string of the task name
         1: a list of dicts containing the parameter definitions
    """

    out = []
    dev_path = os.path.join(package_root, 'dev')
    for fname in sorted(os.listdir(dev_path)):
        if fname.startswith('.') or fname.startswith('_'):
            continue
        if not fname.endswith('.py'):
            continue
        name = fname[:-3]
        args = ()

        full_path = os.path.join(package_root, 'dev', fname)
        with open(full_path, 'rb') as f:
            full_code = f.read()
            if sys.version_info >= (3,):
                full_code = full_code.decode('utf-8')

        task_node = ast.parse(full_code, filename=full_path)
        for node in ast.iter_child_nodes(task_node):
            if isinstance(node, _ast.Assign):
                if len(node.targets) == 1 \
                        and isinstance(node.targets[0], _ast.Name) \
                        and node.targets[0].id == 'run_args':
                    args = ast.literal_eval(node.value)
                    break

        out.append((name, args))
    return out


def show_usage():
    """
    Prints to stderr the valid options for invoking tasks
    """

    valid_tasks = []
    for task in _list_tasks():
        usage = task[0]
        for run_arg in task[1]:
            usage += ' '
            name = run_arg.get('name', '')
            if run_arg.get('required', False):
                usage += '{%s}' % name
            else:
                usage += '[%s]' % name
        valid_tasks.append(usage)

    out = 'Usage: run.py'
    for karg in task_keyword_args:
        out += ' [%s=%s]' % (karg['name'], karg['placeholder'])
    out += ' (%s)' % ' | '.join(valid_tasks)

    print(out, file=sys.stderr)
    sys.exit(1)


def _get_arg(num):
    """
    :return:
        A unicode string of the requested command line arg
    """

    if len(sys.argv) < num + 1:
        return None
    arg = sys.argv[num]
    if isinstance(arg, byte_cls):
        arg = arg.decode('utf-8')
    return arg


def run_task():
    """
    Parses the command line args, invoking the requested task
    """

    arg_num = 1
    task = None
    args = []
    kwargs = {}

    # We look for the task name, processing any global task keyword args
    # by setting the appropriate env var
    while True:
        val = _get_arg(arg_num)
        if val is None:
            break

        next_arg = False
        for karg in task_keyword_args:
            if val.startswith(karg['name'] + '='):
                os.environ[karg['env_var']] = val[len(karg['name']) + 1:]
                next_arg = True
                break

        if next_arg:
            arg_num += 1
            continue

        task = val
        break

    if task is None:
        show_usage()

    task_mod = _import_from('dev.%s' % task, package_root, allow_error=True)
    if task_mod is None:
        show_usage()

    run_args = task_mod.__dict__.get('run_args', [])
    max_args = arg_num + 1 + len(run_args)

    if len(sys.argv) > max_args:
        show_usage()

    for i, run_arg in enumerate(run_args):
        val = _get_arg(arg_num + 1 + i)
        if val is None:
            if run_arg.get('required', False):
                show_usage()
            break

        if run_arg.get('cast') == 'int' and val.isdigit():
            val = int(val)

        kwarg = run_arg.get('kwarg')
        if kwarg:
            kwargs[kwarg] = val
        else:
            args.append(val)

    run = task_mod.__dict__.get('run')

    result = run(*args, **kwargs)
    sys.exit(int(not result))



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import imp
import sys
import os

from . import build_root, package_name, package_root

if sys.version_info < (3,):
    getcwd = os.getcwdu
else:
    getcwd = os.getcwd


def _import_from(mod, path, mod_dir=None, allow_error=False):
    """
    Imports a module from a specific path

    :param mod:
        A unicode string of the module name

    :param path:
        A unicode string to the directory containing the module

    :param mod_dir:
        If the sub directory of "path" is different than the "mod" name,
        pass the sub directory as a unicode string

    :param allow_error:
        If an ImportError should be raised when the module can't be imported

    :return:
        None if not loaded, otherwise the module
    """

    if mod_dir is None:
        mod_dir = mod.replace('.', os.sep)

    if not os.path.exists(path):
        return None

    if not os.path.exists(os.path.join(path, mod_dir)) \
            and not os.path.exists(os.path.join(path, mod_dir + '.py')):
        return None

    if os.sep in mod_dir:
        append, mod_dir = mod_dir.rsplit(os.sep, 1)
        path = os.path.join(path, append)

    try:
        mod_info = imp.find_module(mod_dir, [path])
        return imp.load_module(mod, *mod_info)
    except ImportError:
        if allow_error:
            raise
        return None


def _preload(require_oscrypto, print_info):
    """
    Preloads asn1crypto and optionally oscrypto from a local source checkout,
    or from a normal install

    :param require_oscrypto:
        A bool if oscrypto needs to be preloaded

    :param print_info:
        A bool if info about asn1crypto and oscrypto should be printed
    """

    if print_info:
        print('Working dir: ' + getcwd())
        print('Python ' + sys.version.replace('\n', ''))

    asn1crypto = None
    oscrypto = None

    if require_oscrypto:
        # Some CI services don't use the package name for the dir
        if package_name == 'oscrypto':
            oscrypto_dir = package_root
        else:
            oscrypto_dir = os.path.join(build_root, 'oscrypto')
        oscrypto_tests = None
        if os.path.exists(oscrypto_dir):
            oscrypto_tests = _import_from('oscrypto_tests', oscrypto_dir, 'tests')
        if oscrypto_tests is None:
            import oscrypto_tests
        asn1crypto, oscrypto = oscrypto_tests.local_oscrypto()

    else:
        if package_name == 'asn1crypto':
            asn1crypto_dir = package_root
        else:
            asn1crypto_dir = os.path.join(build_root, 'asn1crypto')
        if os.path.exists(asn1crypto_dir):
            asn1crypto = _import_from('asn1crypto', asn1crypto_dir)
        if asn1crypto is None:
            import asn1crypto

    if print_info:
        print(
            '\nasn1crypto: %s, %s' % (
                asn1crypto.__version__,
                os.path.dirname(asn1crypto.__file__)
            )
        )
        if require_oscrypto:
            backend = oscrypto.backend()
            if backend == 'openssl':
                from oscrypto._openssl._libcrypto import libcrypto_version
                backend = '%s (%s)' % (backend, libcrypto_version)

            print(
                'oscrypto: %s, %s backend, %s' % (
                    oscrypto.__version__,
                    backend,
                    os.path.dirname(oscrypto.__file__)
                )
            )



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os


package_name = "oscrypto"

other_packages = [
    "certbuilder",
    "certvalidator",
    "crlbuilder",
    "csrbuilder",
    "ocspbuilder"
]

task_keyword_args = [
    {
        'name': 'use_openssl',
        'placeholder': '/path/to/libcrypto,/path/to/libssl',
        'env_var': 'OSCRYPTO_USE_OPENSSL',
    },
    {
        'name': 'use_winlegacy',
        'placeholder': 'true',
        'env_var': 'OSCRYPTO_USE_WINLEGACY',
    },
    {
        'name': 'use_ctypes',
        'placeholder': 'true',
        'env_var': 'OSCRYPTO_USE_CTYPES',
    },
    {
        'name': 'skip_internet',
        'placeholder': 'true',
        'env_var': 'OSCRYPTO_SKIP_INTERNET_TESTS',
    },
]

requires_oscrypto = True
has_tests_package = True

package_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
build_root = os.path.abspath(os.path.join(package_root, '..'))

md_source_map = {
    'docs/oscrypto.md': ['oscrypto/__init__.py'],
    'docs/asymmetric.md': ['oscrypto/asymmetric.py', 'oscrypto/_openssl/asymmetric.py'],
    'docs/kdf.md': ['oscrypto/kdf.py', 'oscrypto/_openssl/util.py'],
    'docs/keys.md': ['oscrypto/keys.py', 'oscrypto/_asymmetric.py', 'oscrypto/_openssl/asymmetric.py'],
    'docs/symmetric.md': ['oscrypto/_openssl/symmetric.py'],
    'docs/tls.md': ['oscrypto/tls.py', 'oscrypto/_openssl/tls.py'],
    'docs/trust_list.md': ['oscrypto/trust_list.py'],
    'docs/util.md': ['oscrypto/util.py', 'oscrypto/_rand.py'],
}

definition_replacements = {
    ' is returned by OpenSSL': ' is returned by the OS crypto library'
}



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import imp
import os
import unittest


__version__ = '1.3.0'
__version_info__ = (1, 3, 0)


_asn1crypto_module = None
_oscrypto_module = None


def local_oscrypto():
    """
    Make sure oscrypto is initialized and the backend is selected via env vars

    :return:
        A 2-element tuple with the (asn1crypto, oscrypto) modules
    """

    global _asn1crypto_module
    global _oscrypto_module

    if _oscrypto_module:
        return (_asn1crypto_module, _oscrypto_module)

    tests_dir = os.path.dirname(os.path.abspath(__file__))

    # If we are in a source checkout, load the local oscrypto module, and
    # local asn1crypto module if possible. Otherwise do a normal import.
    in_source_checkout = os.path.basename(tests_dir) == 'tests'

    if in_source_checkout:
        _asn1crypto_module = _import_from(
            'asn1crypto',
            os.path.abspath(os.path.join(tests_dir, '..', '..', 'asn1crypto'))
        )
    if _asn1crypto_module is None:
        import asn1crypto as _asn1crypto_module

    if in_source_checkout:
        _oscrypto_module = _import_from(
            'oscrypto',
            os.path.abspath(os.path.join(tests_dir, '..'))
        )
    if _oscrypto_module is None:
        import oscrypto as _oscrypto_module

    if os.environ.get('OSCRYPTO_USE_CTYPES'):
        _oscrypto_module.use_ctypes()

    # Configuring via env vars so CI for other packages doesn't need to do
    # anything complicated to get the alternate backends
    if os.environ.get('OSCRYPTO_USE_OPENSSL'):
        paths = os.environ.get('OSCRYPTO_USE_OPENSSL').split(',')
        if len(paths) != 2:
            raise ValueError('Value for OSCRYPTO_USE_OPENSSL env var must be two paths separated by a comma')
        _oscrypto_module.use_openssl(*paths)
    elif os.environ.get('OSCRYPTO_USE_WINLEGACY'):
        _oscrypto_module.use_winlegacy()

    return (_asn1crypto_module, _oscrypto_module)


def _import_from(mod, path, mod_dir=None):
    """
    Imports a module from a specific path

    :param mod:
        A unicode string of the module name

    :param path:
        A unicode string to the directory containing the module

    :param mod_dir:
        If the sub directory of "path" is different than the "mod" name,
        pass the sub directory as a unicode string

    :return:
        None if not loaded, otherwise the module
    """

    if mod_dir is None:
        mod_dir = mod

    if not os.path.exists(path):
        return None

    if not os.path.exists(os.path.join(path, mod_dir)):
        return None

    try:
        mod_info = imp.find_module(mod_dir, [path])
        return imp.load_module(mod, *mod_info)
    except ImportError:
        return None


def make_suite():
    """
    Constructs a unittest.TestSuite() of all tests for the package. For use
    with setuptools.

    :return:
        A unittest.TestSuite() object
    """

    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    for test_class in test_classes():
        tests = loader.loadTestsFromTestCase(test_class)
        suite.addTests(tests)
    return suite


def test_classes():
    """
    Returns a list of unittest.TestCase classes for the package

    :return:
        A list of unittest.TestCase classes
    """

    _, oscrypto = local_oscrypto()

    if oscrypto.__version__ != __version__:
        raise AssertionError(
            ('oscrypto_tests version %s can not be run with ' % __version__) +
            ('oscrypto version %s' % oscrypto.__version__)
        )

    from .test_kdf import KDFTests
    from .test_keys import KeyTests
    from .test_asymmetric import AsymmetricTests
    from .test_symmetric import SymmetricTests
    from .test_trust_list import TrustListTests
    from .test_init import InitTests

    test_classes = [
        KDFTests,
        KeyTests,
        AsymmetricTests,
        SymmetricTests,
        TrustListTests,
        InitTests,
    ]
    if not os.environ.get('OSCRYPTO_SKIP_INTERNET_TESTS'):
        from .test_tls import TLSTests
        test_classes.append(TLSTests)

    return test_classes



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import unittest
import sys
import os
import re
import threading
import platform
import time
import socket

from oscrypto import tls, errors, backend
from oscrypto._tls import parse_tls_records, parse_handshake_messages
from asn1crypto import x509

from .unittest_data import data_decorator, data
from .exception_context import assert_exception
from ._unittest_compat import patch
from ._https_client import HttpsClient
from ._socket_proxy import make_socket_proxy
from ._socket_server import make_socket_server

if sys.version_info < (3,):
    import thread
else:
    import _thread as thread

patch()

if sys.version_info < (3,):
    str_cls = unicode  # noqa
    byte_cls = str
else:
    str_cls = str
    byte_cls = bytes


_backend = backend()

xp = sys.platform == 'win32' and sys.getwindowsversion()[0] < 6

tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')

digicert_ca_path = os.path.join(fixtures_dir, 'digicert_ca.crt')
badtls_ca_path = os.path.join(fixtures_dir, 'badtls.io_ca.crt')

# PyPy <= 5.6.0 on OS X 10.11 has a bug with _get_clocktime
osx_pypy_bug = platform.python_implementation() == 'PyPy' \
    and sys.platform == 'darwin' \
    and tuple(map(int, platform.mac_ver()[0].split('.'))) < (10, 12)

raise_with = None
if sys.version_info < (3,):
    exec('''
def raise_with(e, tb):
    raise e, None, tb
''')
else:
    exec('''
def raise_with(e, tb):
    e.__traceback__ = tb
    raise e from None
''')


def connection_timeout(timeout=30):
    def timeout_decorator(f):
        def wrapped(*args):
            try:
                if not osx_pypy_bug:
                    t = threading.Timer(timeout, lambda: thread.interrupt_main())
                    t.start()
                f(*args)
            except (KeyboardInterrupt):
                raise_with(AssertionError("Timed out"), sys.exc_info()[2])
            finally:
                if not osx_pypy_bug:
                    t.cancel()
        return wrapped
    return timeout_decorator


@data_decorator
class TLSTests(unittest.TestCase):

    @staticmethod
    def tls_hosts():
        return (
            ('google', 'www.google.com', 443),
            ('package_control', 'packagecontrol.io', 443),
            ('dh1024', 'dh1024.badtls.io', 10005),
        )

    @data('tls_hosts', True)
    @connection_timeout()
    def tls_connect(self, hostname, port):
        session = None
        if hostname == 'dh1024.badtls.io':
            session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        connection = tls.TLSSocket(hostname, port, session=session)
        self.assertEqual(hostname, connection.hostname)
        self.assertIsInstance(connection.hostname, str_cls)
        self.assertIsInstance(connection.cipher_suite, str_cls)
        self.assertIsInstance(connection.certificate, x509.Certificate)
        self.assertLess(10, len(connection.cipher_suite))
        self.assertEqual(port, connection.port)
        connection.write(b'GET / HTTP/1.1\r\nHost: ' + hostname.encode('utf-8') + b'\r\n\r\n')
        html = connection.read_until(re.compile(b'</html>', re.I))
        self.assertNotEqual(None, re.search(b'</html>', html, re.I))

    @connection_timeout()
    def test_tls_connect_revoked(self):
        if _backend == 'mac':
            from oscrypto._mac._security import osx_version_info

            # macOS 10.12 will read the OCSP Staple response and raise a revoked
            # error, even though we have revocation checking disabled
            if osx_version_info >= (10, 12):
                with assert_exception(self, errors.TLSError, 'revoked'):
                    tls.TLSSocket('global-root-ca-revoked.chain-demos.digicert.com', 443)
                return
        tls.TLSSocket('global-root-ca-revoked.chain-demos.digicert.com', 443)

    @connection_timeout()
    def test_tls_error_http(self):
        with assert_exception(self, errors.TLSError, 'server responded using HTTP'):
            tls.TLSSocket('www.google.com', 80)

    @connection_timeout()
    def test_tls_error_ftp(self):
        s = None
        try:
            def_timeout = socket.getdefaulttimeout()
            socket.setdefaulttimeout(5)
            s = make_socket_server(18021, lambda s, d: s.send(b'220 Welcome to FooFTP\n'))
            with assert_exception(self, errors.TLSError, 'remote end closed the connection|server responded using FTP'):
                tls.TLSSocket('localhost', 18021)
        finally:
            if s:
                s.close()
            socket.setdefaulttimeout(def_timeout)

    @connection_timeout()
    def test_tls_error_missing_issuer(self):
        expected = 'certificate issuer not found in trusted root certificate store'
        with assert_exception(self, errors.TLSVerificationError, expected):
            tls.TLSSocket('domain-match.badtls.io', 10000)

    @connection_timeout()
    def test_tls_error_domain_mismatch(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSVerificationError, 'does not match'):
            tls.TLSSocket('domain-mismatch.badtls.io', 11002, session=session)

    @connection_timeout()
    def test_tls_error_san_mismatch(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSVerificationError, 'does not match'):
            tls.TLSSocket('san-mismatch.badtls.io', 11003, session=session)

    @connection_timeout()
    def test_tls_wildcard_success(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        tls.TLSSocket('wildcard-match.badtls.io', 10001, session=session)

    @connection_timeout()
    def test_tls_error_not_yet_valid(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSVerificationError, 'not valid until'):
            tls.TLSSocket('future.badtls.io', 11001, session=session)

    @connection_timeout()
    def test_tls_error_expired_2(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        # This test allows past or future since cert is 1963, which some systems
        # will interpret as 2063
        with assert_exception(self, errors.TLSVerificationError, 'certificate expired|not valid until'):
            tls.TLSSocket('expired-1963.badtls.io', 11000, session=session)

    @connection_timeout()
    def test_tls_error_client_cert_required(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSError, 'client authentication'):
            tls.TLSSocket('required-auth.badtls.io', 10003, session=session)

    @connection_timeout()
    def test_tls_error_handshake_error_3(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSError, 'weak certificate signature algorithm'):
            tls.TLSSocket('weak-sig.badtls.io', 11004, session=session)

    @connection_timeout()
    def test_tls_error_non_web(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSVerificationError, 'verification failed'):
            tls.TLSSocket('bad-key-usage.badtls.io', 11005, session=session)

    @connection_timeout()
    def test_tls_error_wildcard_mismatch(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSVerificationError, 'does not match'):
            tls.TLSSocket('wildcard.mismatch.badtls.io', 11007, session=session)

    @connection_timeout()
    def test_tls_error_expired(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSVerificationError, 'certificate expired'):
            tls.TLSSocket('expired.badtls.io', 11006, session=session)

    @connection_timeout()
    def test_tls_error_self_signed(self):
        with assert_exception(self, errors.TLSVerificationError, 'self-signed'):
            tls.TLSSocket('self-signed.badssl.com', 443)

    @connection_timeout()
    def test_tls_error_weak_dh_params(self):
        # badssl.com uses SNI, which Windows XP does not support
        regex = 'weak DH parameters' if not xp else 'self-signed'
        # ideally we would use badtls.io since that does not require SNI, however
        # it is not possible to force a good version of OpenSSL to use such a
        # small value for DH params, and I don't feel like the headache of trying
        # to get an old, staticly-linked socat set up just for this text on XP
        with assert_exception(self, errors.TLSError, regex):
            tls.TLSSocket('dh512.badssl.com', 443)

    @connection_timeout()
    def test_tls_error_handshake_error(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSError, 'TLS handshake failed'):
            tls.TLSSocket('rc4-md5.badtls.io', 11009, session=session)

    @connection_timeout()
    def test_tls_error_handshake_error_2(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path])
        with assert_exception(self, errors.TLSError, 'TLS handshake failed'):
            tls.TLSSocket('rc4.badtls.io', 11008, session=session)

    @connection_timeout()
    def test_tls_extra_trust_roots_no_match(self):
        expected = 'certificate issuer not found in trusted root certificate store'
        with assert_exception(self, errors.TLSVerificationError, expected):
            session = tls.TLSSession(extra_trust_roots=[digicert_ca_path])
            tls.TLSSocket('domain-match.badtls.io', 10000, session=session)

    @connection_timeout()
    def test_tls_extra_trust_roots(self):
        session = tls.TLSSession(extra_trust_roots=[badtls_ca_path, digicert_ca_path])
        tls.TLSSocket('domain-match.badtls.io', 10000, session=session)

    @connection_timeout(90)
    def test_tls_graceful_disconnect(self):
        # This connects to a bunch of pages on the same domains using
        # the Connection: Keep-Alive header, however the server is
        # configured to disconnect after 40 responses, so it requires
        # TLSSocket() to properly handle a graceful disconnect.
        urls = [
            'https://packagecontrol.io/browse/popular.json?page=1',
            'https://packagecontrol.io/browse/popular.json?page=2',
            'https://packagecontrol.io/browse/popular.json?page=3',
            'https://packagecontrol.io/browse/popular.json?page=4',
            'https://packagecontrol.io/browse/popular.json?page=5',
            'https://packagecontrol.io/browse/popular.json?page=6',
            'https://packagecontrol.io/browse/popular.json?page=7',
            'https://packagecontrol.io/browse/popular.json?page=8',
            'https://packagecontrol.io/browse/popular.json?page=9',
            'https://packagecontrol.io/browse/popular.json?page=10',
            'https://packagecontrol.io/browse/popular.json?page=11',
            'https://packagecontrol.io/browse/popular.json?page=12',
            'https://packagecontrol.io/browse/popular.json?page=13',
            'https://packagecontrol.io/browse/popular.json?page=14',
            'https://packagecontrol.io/browse/popular.json?page=15',
            'https://packagecontrol.io/browse/popular.json?page=16',
            'https://packagecontrol.io/browse/popular.json?page=17',
            'https://packagecontrol.io/browse/popular.json?page=18',
            'https://packagecontrol.io/browse/popular.json?page=19',
            'https://packagecontrol.io/browse/popular.json?page=20',
            'https://packagecontrol.io/browse/popular.json?page=21',
            'https://packagecontrol.io/browse/popular.json?page=22',
            'https://packagecontrol.io/browse/popular.json?page=23',
            'https://packagecontrol.io/browse/popular.json?page=24',
            'https://packagecontrol.io/browse/popular.json?page=25',
            'https://packagecontrol.io/browse/popular.json?page=26',
            'https://packagecontrol.io/browse/popular.json?page=27',
            'https://packagecontrol.io/browse/popular.json?page=28',
            'https://packagecontrol.io/browse/popular.json?page=29',
            'https://packagecontrol.io/browse/popular.json?page=30',
            'https://packagecontrol.io/browse/popular.json?page=31',
            'https://packagecontrol.io/browse/popular.json?page=32',
            'https://packagecontrol.io/browse/popular.json?page=33',
            'https://packagecontrol.io/browse/popular.json?page=34',
            'https://packagecontrol.io/browse/popular.json?page=35',
            'https://packagecontrol.io/browse/popular.json?page=36',
            'https://packagecontrol.io/browse/popular.json?page=37',
            'https://packagecontrol.io/browse/popular.json?page=38',
            'https://packagecontrol.io/browse/popular.json?page=39',
            'https://packagecontrol.io/browse/popular.json?page=40',
            'https://packagecontrol.io/browse/popular.json?page=41',
            'https://packagecontrol.io/browse/popular.json?page=42',
            'https://packagecontrol.io/browse/popular.json?page=43',
            'https://packagecontrol.io/browse/popular.json?page=44',
            'https://packagecontrol.io/browse/popular.json?page=45',
            'https://packagecontrol.io/browse/popular.json?page=46',
            'https://packagecontrol.io/browse/popular.json?page=47',
            'https://packagecontrol.io/browse/popular.json?page=48',
            'https://packagecontrol.io/browse/popular.json?page=49',
            'https://packagecontrol.io/browse/popular.json?page=50',
            'https://packagecontrol.io/browse/popular.json?page=51',
            'https://packagecontrol.io/browse/popular.json?page=52',
            'https://packagecontrol.io/browse/popular.json?page=53',
            'https://packagecontrol.io/browse/popular.json?page=54',
            'https://packagecontrol.io/browse/popular.json?page=55',
            'https://packagecontrol.io/browse/popular.json?page=56',
            'https://packagecontrol.io/browse/popular.json?page=57',
            'https://packagecontrol.io/browse/popular.json?page=58',
            'https://packagecontrol.io/browse/popular.json?page=59',
            'https://packagecontrol.io/browse/popular.json?page=60',
            'https://packagecontrol.io/browse/popular.json?page=61',
            'https://packagecontrol.io/browse/popular.json?page=62',
            'https://packagecontrol.io/browse/popular.json?page=63',
            'https://packagecontrol.io/browse/popular.json?page=64',
            'https://packagecontrol.io/browse/popular.json?page=65',
            'https://packagecontrol.io/browse/popular.json?page=66',
            'https://packagecontrol.io/browse/popular.json?page=67',
            'https://packagecontrol.io/browse/popular.json?page=68',
            'https://packagecontrol.io/browse/popular.json?page=69',
            'https://packagecontrol.io/browse/popular.json?page=70',
            'https://packagecontrol.io/browse/popular.json?page=71',
            'https://packagecontrol.io/browse/popular.json?page=72',
            'https://packagecontrol.io/browse/popular.json?page=73',
            'https://packagecontrol.io/browse/popular.json?page=74',
            'https://packagecontrol.io/browse/popular.json?page=75',
            'https://packagecontrol.io/browse/popular.json?page=76',
            'https://packagecontrol.io/browse/popular.json?page=77',
            'https://packagecontrol.io/browse/popular.json?page=78',
            'https://packagecontrol.io/browse/popular.json?page=79',
            'https://packagecontrol.io/browse/popular.json?page=80',
            'https://packagecontrol.io/browse/popular.json?page=81',
            'https://packagecontrol.io/browse/popular.json?page=82',
            'https://packagecontrol.io/browse/popular.json?page=83',
            'https://packagecontrol.io/browse/popular.json?page=84',
            'https://packagecontrol.io/browse/popular.json?page=85',
            'https://packagecontrol.io/browse/popular.json?page=86',
            'https://packagecontrol.io/browse/popular.json?page=87',
            'https://packagecontrol.io/browse/popular.json?page=88',
            'https://packagecontrol.io/browse/popular.json?page=89',
            'https://packagecontrol.io/browse/popular.json?page=90',
            'https://packagecontrol.io/browse/popular.json?page=91',
            'https://packagecontrol.io/browse/popular.json?page=92',
            'https://packagecontrol.io/browse/popular.json?page=93',
            'https://packagecontrol.io/browse/popular.json?page=94',
            'https://packagecontrol.io/browse/popular.json?page=95',
            'https://packagecontrol.io/browse/popular.json?page=96',
            'https://packagecontrol.io/browse/popular.json?page=97',
            'https://packagecontrol.io/browse/popular.json?page=98',
            'https://packagecontrol.io/browse/popular.json?page=99',
            'https://packagecontrol.io/browse/popular.json?page=100',
        ]
        c = HttpsClient()
        for url in urls:
            c.download(url, 15)

    @connection_timeout(60)
    def test_tls_large_download(self):
        # This tests downloading a large (3MB+) file to ensure
        # there aren't buffer overflow issues in TLSSocket()
        c = HttpsClient()
        c.download('https://packagecontrol.io/channel_v3.json', 15)

    @connection_timeout()
    def test_tls_protocol_version(self):
        session = tls.TLSSession(set(['TLSv1', 'TLSv1.1']))
        with assert_exception(self, errors.TLSError, 'TLS handshake failed - protocol version error'):
            tls.TLSSocket('github.com', 443, session=session)

    @connection_timeout()
    def test_tls_pause_timeout(self):
        c = HttpsClient(False, True)
        c.download('https://packagecontrol.io/browse/popular.json?page=1', 5)
        time.sleep(10)
        c.download('https://packagecontrol.io/browse/popular.json?page=2', 5)

    @connection_timeout()
    def test_tls_closed_connection_read(self):
        ip = socket.gethostbyname('badtls.io')

        with assert_exception(self, (errors.TLSDisconnectError, errors.TLSError),
                              'The remote end closed the connection|TLS handshake failed'):
            try:
                sock, send_sock, recv_sock, server = make_socket_proxy(ip, 443)
                tsock = tls.TLSSocket.wrap(sock, 'badtls.io')
                send_sock.close()
                tsock.read(8192)
            finally:
                recv_sock.close()
                server.close()

    @connection_timeout()
    def test_tls_closed_connection_write(self):
        ip = socket.gethostbyname('badtls.io')

        with assert_exception(self, (errors.TLSDisconnectError, errors.TLSError),
                              'The remote end closed the connection|TLS handshake failed'):
            try:
                sock, send_sock, recv_sock, server = make_socket_proxy(ip, 443)
                tsock = tls.TLSSocket.wrap(sock, 'badtls.io')
                send_sock.close()
                tsock.write(b'GET / HTTP/1.1\r\n')
                tsock.write(b'\r\n')
                tsock.write(b'\r\n')
            finally:
                recv_sock.close()
                server.close()

    @connection_timeout()
    def test_tls_closed_connection_read_handshake(self):
        ip = socket.gethostbyname('badtls.io')

        # Break the connection after the ServerHello is received
        def recv_callback(src, dest):
            data = src.recv(8192)
            for record_type, tls_version, message in parse_tls_records(data):
                if record_type == b'\x16':
                    for message_type, payload in parse_handshake_messages(message):
                        if message_type == b'\x02':
                            dest.close()
                            return
            dest.send(data)

        with assert_exception(self, (errors.TLSDisconnectError, errors.TLSError),
                              'The remote end closed the connection|TLS handshake failed'):
            try:
                sock, send_sock, recv_sock, server = make_socket_proxy(ip, 443, None, recv_callback)
                tls.TLSSocket.wrap(sock, 'badtls.io')
            finally:
                recv_sock.close()
                send_sock.close()
                server.close()

    @connection_timeout()
    def test_tls_closed_connection_write_handshake(self):
        ip = socket.gethostbyname('badtls.io')

        # Break the connection after the ClientHello is sent
        def send_callback(src, dest):
            data = src.recv(8192)
            for record_type, tls_version, message in parse_tls_records(data):
                if record_type == b'\x16':
                    for message_type, payload in parse_handshake_messages(message):
                        if message_type == b'\x01':
                            src.close()
                            return
            dest.send(data)

        with assert_exception(self, (errors.TLSDisconnectError, errors.TLSError),
                              'The remote end closed the connection|TLS handshake failed'):
            recv_sock = None
            send_sock = None
            server = None
            try:
                sock, send_sock, recv_sock, server = make_socket_proxy(ip, 443, send_callback)
                tls.TLSSocket.wrap(sock, 'badtls.io')
            finally:
                if recv_sock:
                    recv_sock.close()
                if send_sock:
                    send_sock.close()
                if server:
                    server.close()

    @connection_timeout()
    def test_tls_closed_connection_read_shutdown(self):
        ip = socket.gethostbyname('badtls.io')
        try:
            sock, send_sock, recv_sock, server = make_socket_proxy(ip, 443)
            tsock = None
            try:
                tsock = tls.TLSSocket.wrap(sock, 'badtls.io')
                send_sock.close()
                tsock.read(8192)
                shutdown = False
            except (errors.TLSError):
                if tsock:
                    tsock.shutdown()
                shutdown = True
            self.assertEqual(True, shutdown)
        finally:
            recv_sock.close()
            server.close()

    @connection_timeout()
    def test_tls_closed_connection_write_shutdown(self):
        ip = socket.gethostbyname('badtls.io')
        try:
            sock, send_sock, recv_sock, server = make_socket_proxy(ip, 443)
            tsock = None
            try:
                tsock = tls.TLSSocket.wrap(sock, 'badtls.io')
                send_sock.close()
                tsock.write(b'GET / HTTP/1.1\r\n')
                tsock.write(b'\r\n')
                tsock.write(b'\r\n')
                tsock.read(8192)
                shutdown = False
            except (errors.TLSError):
                if tsock:
                    tsock.shutdown()
                shutdown = True
            self.assertEqual(True, shutdown)
        finally:
            recv_sock.close()
            server.close()



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import re
import sys
from contextlib import contextmanager


if sys.version_info < (3,):
    str_cls = unicode  # noqa
else:
    str_cls = str


@contextmanager
def assert_exception(test_case, expected_class, expected_msg):
    """
    Look for a specific exception type and message, allowing the exception
    to be raised if it doesn't match
    """

    expected_re = re.compile(expected_msg)

    try:
        yield

    except Exception as e:
        should_raise = True
        if isinstance(e, expected_class):
            test_case.assertIsInstance(e, expected_class)
            msg = str_cls(e)
            if expected_re.search(msg):
                test_case.assertRegex(msg, expected_re)
                should_raise = False
        if should_raise:
            raise



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import unittest
import sys
import os

from asn1crypto import pem, algos, keys, core
from oscrypto import asymmetric, errors, backend

from ._unittest_compat import patch

patch()

if sys.version_info < (3,):
    byte_cls = str
    int_types = (int, long)  # noqa
else:
    byte_cls = bytes
    int_types = (int,)


_backend = backend()


if _backend == 'openssl':
    from oscrypto._openssl._libcrypto import libcrypto_version_info
    openssl_098 = libcrypto_version_info < (1, 0, 0)
else:
    openssl_098 = False


tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')


def _win_version_pair():
    ver_info = sys.getwindowsversion()
    return (ver_info[0], ver_info[1])


def _should_support_sha2():
    if _backend == 'mac':
        return False
    if _backend == 'winlegacy':
        return False
    if _backend == 'win' and _win_version_pair() < (6, 2):
        return False
    if openssl_098:
        return False
    return True


class AsymmetricTests(unittest.TestCase):

    def test_load_incomplete_dsa_cert(self):
        with self.assertRaises(errors.IncompleteAsymmetricKeyError):
            asymmetric.load_public_key(os.path.join(fixtures_dir, 'DSAParametersInheritedCACert.crt'))

    def test_cert_attributes(self):
        cert = asymmetric.load_certificate(os.path.join(fixtures_dir, 'keys/test.crt'))
        self.assertEqual(2048, cert.bit_size)
        self.assertEqual(256, cert.byte_size)
        self.assertEqual('rsa', cert.algorithm)

    def test_public_key_attributes(self):
        pub_key = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-public-rsa.key'))
        self.assertEqual(2048, pub_key.bit_size)
        self.assertEqual(256, pub_key.byte_size)
        self.assertEqual('rsa', pub_key.algorithm)

    def test_private_key_attributes(self):
        private_key = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        self.assertEqual(2048, private_key.bit_size)
        self.assertEqual(256, private_key.byte_size)
        self.assertEqual('rsa', private_key.algorithm)

    def test_cert_ec_attributes(self):
        cert = asymmetric.load_certificate(os.path.join(fixtures_dir, 'keys/test-ec-named.crt'))
        self.assertEqual(256, cert.bit_size)
        self.assertEqual(32, cert.byte_size)
        self.assertEqual('secp256r1', cert.curve)
        self.assertEqual('ec', cert.algorithm)

    def test_public_key_ec_attributes(self):
        pub_key = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-public-ec-named.key'))
        self.assertEqual(256, pub_key.bit_size)
        self.assertEqual(32, pub_key.byte_size)
        self.assertEqual('secp256r1', pub_key.curve)
        self.assertEqual('ec', pub_key.algorithm)

    def test_private_key_ec_attributes(self):
        private_key = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-ec-named.key'))
        self.assertEqual(256, private_key.bit_size)
        self.assertEqual(32, private_key.byte_size)
        self.assertEqual('secp256r1', private_key.curve)
        self.assertEqual('ec', private_key.algorithm)

    def test_dump_public(self):
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        pem_serialized = asymmetric.dump_public_key(public)
        public_reloaded = asymmetric.load_public_key(pem_serialized)
        self.assertIsInstance(public_reloaded, asymmetric.PublicKey)
        self.assertEqual('rsa', public_reloaded.algorithm)

    def test_dump_certificate(self):
        cert = asymmetric.load_certificate(os.path.join(fixtures_dir, 'keys/test.crt'))
        pem_serialized = asymmetric.dump_certificate(cert)
        cert_reloaded = asymmetric.load_certificate(pem_serialized)
        self.assertIsInstance(cert_reloaded, asymmetric.Certificate)
        self.assertEqual('rsa', cert_reloaded.algorithm)

    def test_dump_private(self):
        def do_run():
            private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))

            for password in [None, 'password123']:
                pem_serialized = asymmetric.dump_private_key(private, password, target_ms=20)
                private_reloaded = asymmetric.load_private_key(pem_serialized, password)
                self.assertTrue(pem.detect(pem_serialized))
                self.assertIsInstance(private_reloaded, asymmetric.PrivateKey)
                self.assertEqual('rsa', private_reloaded.algorithm)

        # OpenSSL 0.9.8 and Windows CryptoAPI don't have PBKDF2 implemented in
        # C, thus the dump operation fails since there is no reasonable way to
        # ensure we are using a good number of iterations of PBKDF2
        if openssl_098 or _backend == 'winlegacy':
            with self.assertRaises(OSError):
                do_run()
        else:
            do_run()

    def test_dump_private_openssl(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        pem_serialized = asymmetric.dump_openssl_private_key(private, 'password123')
        private_reloaded = asymmetric.load_private_key(pem_serialized, 'password123')
        self.assertIsInstance(private_reloaded, asymmetric.PrivateKey)
        self.assertEqual('rsa', private_reloaded.algorithm)

    def test_load_rsa_pss_cert(self):
        cert = asymmetric.load_certificate(os.path.join(fixtures_dir, 'keys/test-pss.crt'))
        self.assertEqual('rsassa_pss', cert.algorithm)
        self.assertEqual(2048, cert.bit_size)

    def test_dh_generate(self):
        dh_parameters = asymmetric.generate_dh_parameters(512)
        self.assertIsInstance(dh_parameters, algos.DHParameters)
        self.assertIsInstance(dh_parameters['p'].native, int_types)
        self.assertIsInstance(dh_parameters['g'].native, int_types)
        self.assertEqual(2, dh_parameters['g'].native)

    def test_rsa_generate(self):
        public, private = asymmetric.generate_pair('rsa', bit_size=2048)

        self.assertEqual('rsa', public.algorithm)
        self.assertEqual(2048, public.bit_size)

        original_data = b'This is data to sign'
        signature = asymmetric.rsa_pkcs1v15_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)
        asymmetric.rsa_pkcs1v15_verify(public, signature, original_data, 'sha1')

        raw_public = asymmetric.dump_public_key(public)
        asymmetric.load_public_key(raw_public)
        raw_private = asymmetric.dump_private_key(private, None)
        asymmetric.load_private_key(raw_private, None)

        self.assertIsInstance(private.fingerprint, byte_cls)
        self.assertIsInstance(public.fingerprint, byte_cls)
        self.assertEqual(private.fingerprint, public.fingerprint)

    def test_dsa_generate(self):
        public, private = asymmetric.generate_pair('dsa', bit_size=1024)

        self.assertEqual('dsa', public.algorithm)
        self.assertEqual(1024, public.bit_size)

        original_data = b'This is data to sign'
        signature = asymmetric.dsa_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)
        asymmetric.dsa_verify(public, signature, original_data, 'sha1')

        raw_public = asymmetric.dump_public_key(public)
        asymmetric.load_public_key(raw_public)
        raw_private = asymmetric.dump_private_key(private, None)
        asymmetric.load_private_key(raw_private, None)

        self.assertIsInstance(private.fingerprint, byte_cls)
        self.assertIsInstance(public.fingerprint, byte_cls)
        self.assertEqual(private.fingerprint, public.fingerprint)

    def test_ec_generate(self):
        public, private = asymmetric.generate_pair('ec', curve='secp256r1')

        self.assertEqual('ec', public.algorithm)
        self.assertEqual('secp256r1', public.asn1.curve[1])

        original_data = b'This is data to sign'
        signature = asymmetric.ecdsa_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)
        asymmetric.ecdsa_verify(public, signature, original_data, 'sha1')

        raw_public = asymmetric.dump_public_key(public)
        asymmetric.load_public_key(raw_public)
        raw_private = asymmetric.dump_private_key(private, None)
        asymmetric.load_private_key(raw_private, None)

        self.assertIsInstance(private.fingerprint, byte_cls)
        self.assertIsInstance(public.fingerprint, byte_cls)
        self.assertEqual(private.fingerprint, public.fingerprint)

    def test_rsa_verify(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        asymmetric.rsa_pkcs1v15_verify(public, signature, original_data, 'sha1')

    def test_rsa_verify_key_size_mismatch(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-4096.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.rsa_pkcs1v15_verify(public, signature, original_data, 'sha1')

    def test_rsa_verify_fail(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.rsa_pkcs1v15_verify(public, signature, original_data + b'1', 'sha1')

    def test_rsa_verify_fail_each_byte(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_signature'), 'rb') as f:
            original_signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        for i in range(0, len(original_signature)):
            if i == 0:
                signature = b'\xab' + original_signature[1:]
            elif i == len(original_signature) - 1:
                signature = original_signature[0:-1] + b'\xab'
            else:
                signature = original_signature[0:i] + b'\xab' + original_signature[i + 1:]
            with self.assertRaises(errors.SignatureError):
                asymmetric.rsa_pkcs1v15_verify(public, signature, original_data + b'1', 'sha1')

    def test_rsa_pss_verify(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_pss_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        asymmetric.rsa_pss_verify(public, signature, original_data, 'sha1')

    def test_rsa_pss_verify_pss_cert(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_pss_signature_pss_cert'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-pss.crt'))
        asymmetric.rsa_pss_verify(public, signature, original_data, 'sha256')

    def test_rsa_pss_verify_fail(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_pss_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.rsa_pss_verify(public, signature, original_data + b'1', 'sha1')

    def test_rsa_pss_verify_pss_cert_fail(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_pss_signature_pss_cert'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-pss.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.rsa_pss_verify(public, signature, original_data + b'1', 'sha256')

    def test_rsa_raw_verify(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_signature_raw'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        asymmetric.rsa_pkcs1v15_verify(public, signature, original_data, 'raw')

    def test_rsa_raw_verify_fail(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'rsa_signature_raw'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.rsa_pkcs1v15_verify(public, signature, original_data + b'1', 'raw')

    def test_dsa_verify(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'dsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))
        asymmetric.dsa_verify(public, signature, original_data, 'sha1')

    def test_dsa_verify_key_size_mismatch(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'dsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-512.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.dsa_verify(public, signature, original_data, 'sha1')

    def test_dsa_verify_fail(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'dsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))
        with self.assertRaises(errors.SignatureError):
            asymmetric.dsa_verify(public, signature, original_data + b'1', 'sha1')

    def test_dsa_verify_fail_each_byte(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'dsa_signature'), 'rb') as f:
            original_signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))
        for i in range(0, len(original_signature)):
            if i == 0:
                signature = b'\xab' + original_signature[1:]
            elif i == len(original_signature) - 1:
                signature = original_signature[0:-1] + b'\xab'
            else:
                signature = original_signature[0:i] + b'\xab' + original_signature[i+1:]
            with self.assertRaises(errors.SignatureError):
                asymmetric.dsa_verify(public, signature, original_data + b'1', 'sha1')

    def test_ecdsa_verify(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'ecdsa_signature'), 'rb') as f:
            signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-public-ec-named.key'))
        asymmetric.ecdsa_verify(public, signature, original_data, 'sha1')

    def test_ecdsa_verify_fail_each_byte(self):
        with open(os.path.join(fixtures_dir, 'message.txt'), 'rb') as f:
            original_data = f.read()
        with open(os.path.join(fixtures_dir, 'ecdsa_signature'), 'rb') as f:
            original_signature = f.read()
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-public-ec-named.key'))
        for i in range(0, len(original_signature)):
            if i == 0:
                signature = b'\xab' + original_signature[1:]
            elif i == len(original_signature) - 1:
                signature = original_signature[0:-1] + b'\xab'
            else:
                signature = original_signature[0:i] + b'\xab' + original_signature[i+1:]
            with self.assertRaises(errors.SignatureError):
                asymmetric.ecdsa_verify(public, signature, original_data + b'1', 'sha1')

    def test_rsa_pkcs1v15_encrypt(self):
        original_data = b'This is data to encrypt'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        ciphertext = asymmetric.rsa_pkcs1v15_encrypt(public, original_data)
        self.assertIsInstance(ciphertext, byte_cls)

        plaintext = asymmetric.rsa_pkcs1v15_decrypt(private, ciphertext)
        self.assertEqual(original_data, plaintext)

    def test_rsa_oaep_encrypt(self):
        original_data = b'This is data to encrypt'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        ciphertext = asymmetric.rsa_oaep_encrypt(public, original_data)
        self.assertIsInstance(ciphertext, byte_cls)

        plaintext = asymmetric.rsa_oaep_decrypt(private, ciphertext)
        self.assertEqual(original_data, plaintext)

    def test_rsa_private_pkcs1v15_decrypt(self):
        original_data = b'This is the message to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))

        with open(os.path.join(fixtures_dir, 'rsa_public_encrypted'), 'rb') as f:
            plaintext = asymmetric.rsa_pkcs1v15_decrypt(private, f.read())
            self.assertEqual(original_data, plaintext)

    def test_rsa_private_oaep_decrypt(self):
        original_data = b'This is the message to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))

        with open(os.path.join(fixtures_dir, 'rsa_public_encrypted_oaep'), 'rb') as f:
            plaintext = asymmetric.rsa_oaep_decrypt(private, f.read())
            self.assertEqual(original_data, plaintext)

    def test_rsa_sign(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        signature = asymmetric.rsa_pkcs1v15_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.rsa_pkcs1v15_verify(public, signature, original_data, 'sha1')

    def test_rsa_fingerprint(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        self.assertIsInstance(private.fingerprint, byte_cls)
        self.assertIsInstance(public.fingerprint, byte_cls)
        self.assertEqual(private.fingerprint, public.fingerprint)

    def test_rsa_public_key_attr(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        computed_public = private.public_key
        self.assertEqual(public.asn1.dump(), computed_public.asn1.dump())

    def test_rsa_private_key_unwrap(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        self.assertIsInstance(private.unwrap(), keys.RSAPrivateKey)

    def test_rsa_public_key_unwrap(self):
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))
        self.assertIsInstance(public.unwrap(), keys.RSAPublicKey)

    def test_rsa_pss_sign(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        signature = asymmetric.rsa_pss_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.rsa_pss_verify(public, signature, original_data, 'sha1')

    def test_rsa_pss_sign_pss_cert(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-pss.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-pss.crt'))

        signature = asymmetric.rsa_pss_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.rsa_pss_verify(public, signature, original_data, 'sha1')

    def test_rsa_pss_sha256_sign(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        signature = asymmetric.rsa_pss_sign(private, original_data, 'sha256')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.rsa_pss_verify(public, signature, original_data, 'sha256')

    def test_rsa_pss_sha256_sign_pss_cert(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-pss.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-pss.crt'))

        signature = asymmetric.rsa_pss_sign(private, original_data, 'sha256')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.rsa_pss_verify(public, signature, original_data, 'sha256')

    def test_rsa_raw_sign(self):
        original_data = b'This is data to sign!'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test.crt'))

        signature = asymmetric.rsa_pkcs1v15_sign(private, original_data, 'raw')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.rsa_pkcs1v15_verify(public, signature, original_data, 'raw')

    def test_dsa_sign(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))

        signature = asymmetric.dsa_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.dsa_verify(public, signature, original_data, 'sha1')

    def test_dsa_fingerprint(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))

        self.assertIsInstance(private.fingerprint, byte_cls)
        self.assertIsInstance(public.fingerprint, byte_cls)
        self.assertEqual(private.fingerprint, public.fingerprint)

    def test_dsa_public_key_attr(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))

        computed_public = private.public_key
        self.assertEqual(public.asn1.dump(), computed_public.asn1.dump())

    def test_dsa_private_key_unwrap(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.key'))
        self.assertIsInstance(private.unwrap(), keys.DSAPrivateKey)

    def test_dsa_public_key_unwrap(self):
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-1024.crt'))
        self.assertIsInstance(public.unwrap(), core.Integer)

    def test_dsa_2048_sha1_sign(self):
        def do_run():
            original_data = b'This is data to sign'
            private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa-2048.key'))
            public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-2048.crt'))

            signature = asymmetric.dsa_sign(private, original_data, 'sha1')
            self.assertIsInstance(signature, byte_cls)

            asymmetric.dsa_verify(public, signature, original_data, 'sha1')

        if sys.platform == 'win32':
            with self.assertRaises(errors.AsymmetricKeyError):
                do_run()
        else:
            do_run()

    def test_dsa_2048_sha2_sign(self):
        def do_run():
            original_data = b'This is data to sign'
            private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa-2048-sha2.key'))
            public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa-2048-sha2.crt'))

            signature = asymmetric.dsa_sign(private, original_data, 'sha256')
            self.assertIsInstance(signature, byte_cls)

            asymmetric.dsa_verify(public, signature, original_data, 'sha256')

        if not _should_support_sha2():
            with self.assertRaises(errors.AsymmetricKeyError):
                do_run()
        else:
            do_run()

    def test_dsa_3072_sign(self):
        def do_run():
            original_data = b'This is data to sign'
            private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa.key'))
            public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa.crt'))

            signature = asymmetric.dsa_sign(private, original_data, 'sha256')
            self.assertIsInstance(signature, byte_cls)

            asymmetric.dsa_verify(public, signature, original_data, 'sha256')

        if not _should_support_sha2():
            with self.assertRaises(errors.AsymmetricKeyError):
                do_run()
        else:
            do_run()

    def test_dsa_3072_sign_sha1(self):
        def do_run():
            original_data = b'This is data to sign'
            private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-dsa.key'))
            public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-dsa.crt'))

            signature = asymmetric.dsa_sign(private, original_data, 'sha1')
            self.assertIsInstance(signature, byte_cls)

            asymmetric.dsa_verify(public, signature, original_data, 'sha1')

        if _backend == 'mac' or openssl_098 or _backend == 'winlegacy':
            with self.assertRaises(errors.AsymmetricKeyError):
                do_run()
        elif _backend == 'win':
            if _win_version_pair() < (6, 2):
                exception_class = errors.AsymmetricKeyError
            else:
                exception_class = ValueError
            with self.assertRaises(exception_class):
                do_run()
        else:
            do_run()

    def test_ecdsa_sign(self):
        original_data = b'This is data to sign'
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-ec-named.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-ec-named.crt'))

        signature = asymmetric.ecdsa_sign(private, original_data, 'sha1')
        self.assertIsInstance(signature, byte_cls)

        asymmetric.ecdsa_verify(public, signature, original_data, 'sha1')

    def test_ec_fingerprints(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-ec-named.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-ec-named.crt'))

        self.assertIsInstance(private.fingerprint, byte_cls)
        self.assertIsInstance(public.fingerprint, byte_cls)
        self.assertEqual(private.fingerprint, public.fingerprint)

    def test_ec_public_key_attr(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-ec-named.key'))
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-ec-named.crt'))

        computed_public = private.public_key
        self.assertEqual(public.asn1.dump(), computed_public.asn1.dump())

    def test_ec_private_key_unwrap(self):
        private = asymmetric.load_private_key(os.path.join(fixtures_dir, 'keys/test-ec-named.key'))
        self.assertIsInstance(private.unwrap(), keys.ECPrivateKey)

    def test_ec_public_key_unwrap(self):
        public = asymmetric.load_public_key(os.path.join(fixtures_dir, 'keys/test-ec-named.crt'))
        self.assertIsInstance(public.unwrap(), keys.ECPointBitString)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os
import shutil
import subprocess
import sys
from urllib.parse import urlparse
from urllib.request import urlopen


run_args = [
    {
        'name': 'version',
        'kwarg': 'version',
    },
    {
        'name': 'arch',
        'kwarg': 'arch',
    },
]


def run(version=None, arch=None):
    """
    Installs a version of Python on Windows

    :return:
        A bool - if Python was installed successfully
    """

    if sys.platform != 'win32':
        raise ValueError('python-install is only designed for Windows')

    if version not in set(['2.6', '3.3']):
        raise ValueError('Invalid version: %r' % version)

    if arch not in set(['x86', 'x64']):
        raise ValueError('Invalid arch: %r' % arch)

    if version == '2.6':
        if arch == 'x64':
            url = 'https://www.python.org/ftp/python/2.6.6/python-2.6.6.amd64.msi'
        else:
            url = 'https://www.python.org/ftp/python/2.6.6/python-2.6.6.msi'
    else:
        if arch == 'x64':
            url = 'https://www.python.org/ftp/python/3.3.5/python-3.3.5.amd64.msi'
        else:
            url = 'https://www.python.org/ftp/python/3.3.5/python-3.3.5.msi'

    home = os.environ.get('USERPROFILE')
    msi_filename = os.path.basename(urlparse(url).path)
    msi_path = os.path.join(home, msi_filename)
    install_path = os.path.join(os.environ.get('LOCALAPPDATA'), 'Python%s-%s' % (version, arch))

    if os.path.exists(os.path.join(install_path, 'python.exe')):
        print(install_path)
        return True

    try:
        with urlopen(url) as r, open(msi_path, 'wb') as f:
            shutil.copyfileobj(r, f)

        proc = subprocess.Popen(
            'msiexec /passive /a %s TARGETDIR=%s' % (msi_filename, install_path),
            shell=True,
            cwd=home
        )
        proc.communicate()

    finally:
        if os.path.exists(msi_path):
            os.unlink(msi_path)

    print(install_path)
    return True



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import ast
import _ast
import unittest
import os
import sys

import oscrypto as module


# This handles situations where an import is importing a function from a
# dotted path, e.g. "from . import ident", and ident is a function, not a
# submodule
MOD_MAP = {
    'oscrypto._backend_config': 'oscrypto',
    'oscrypto.backend': 'oscrypto',
    'oscrypto.ffi': 'oscrypto'
}


def add_mod(mod_name, imports):
    """
    Maps pre-defined module.function to module import names

    :param mod_name:
        A unicode string of a fully-qualified module name being imported

    :param imports:
        A set of unicode strings of the modules that are being imported
    """

    imports.add(MOD_MAP.get(mod_name, mod_name))


def walk_ast(parent_node, modname, imports):
    """
    Walks the AST for a module finding any imports and recording them

    :param parent_node:
        A node from the _ast module

    :param modname:
        A unicode string of the module we are walking the AST of

    :param imports:
        A set of unicode strings of the imports that have been found so far
    """

    for node in ast.iter_child_nodes(parent_node):
        if isinstance(node, _ast.Import):
            if node.names[0].name.startswith(module.__name__):
                add_mod(node.names[0].name, imports)

        elif isinstance(node, _ast.ImportFrom):
            if node.level > 0:
                if modname == module.__name__:
                    base_mod = module.__name__
                else:
                    base_mod = '.'.join(modname.split('.')[:-node.level])
                if node.module:
                    base_mod += '.' + node.module
            else:
                base_mod = node.module

            if not base_mod.startswith(module.__name__):
                continue

            if node.level > 0 and not node.module:
                for n in node.names:
                    add_mod(base_mod + '.' + n.name, imports)
            else:
                add_mod(base_mod, imports)

        elif isinstance(node, _ast.If):
            for subast in node.body:
                walk_ast(subast, modname, imports)
            for subast in node.orelse:
                walk_ast(subast, modname, imports)

        elif sys.version_info >= (3, 3) and isinstance(node, _ast.Try):
            for subast in node.body:
                walk_ast(subast, modname, imports)
            for subast in node.orelse:
                walk_ast(subast, modname, imports)
            for subast in node.finalbody:
                walk_ast(subast, modname, imports)

        elif sys.version_info < (3, 3) and isinstance(node, _ast.TryFinally):
            for subast in node.body:
                walk_ast(subast, modname, imports)
            for subast in node.finalbody:
                walk_ast(subast, modname, imports)

        elif sys.version_info < (3, 3) and isinstance(node, _ast.TryExcept):
            for subast in node.body:
                walk_ast(subast, modname, imports)
            for subast in node.orelse:
                walk_ast(subast, modname, imports)


class InitTests(unittest.TestCase):

    def test_load_order(self):
        deps = {}

        mod_root = os.path.abspath(os.path.dirname(module.__file__))
        files = []
        for root, dnames, fnames in os.walk(mod_root):
            for f in fnames:
                if f.endswith('.py'):
                    full_path = os.path.join(root, f)
                    rel_path = full_path.replace(mod_root + os.sep, '')
                    files.append((full_path, rel_path))

        for full_path, rel_path in sorted(files):
            with open(full_path, 'rb') as f:
                full_code = f.read()
                if sys.version_info >= (3,):
                    full_code = full_code.decode('utf-8')

            modname = rel_path.replace('.py', '').replace(os.sep, '.')
            if modname == '__init__':
                modname = module.__name__
            else:
                modname = '%s.%s' % (module.__name__, modname)

            if sys.version_info < (3,) and sys.platform == 'win32' and b'\r\n' in full_code:
                full_code = full_code.replace(b'\r\n', b'\n')

            imports = set([])
            module_node = ast.parse(full_code, filename=full_path)
            walk_ast(module_node, modname, imports)

            deps[modname] = imports

        load_order = module.load_order()
        prev = set([])
        for mod in load_order:
            self.assertEqual(True, mod in deps)
            self.assertEqual((mod, set([])), (mod, deps[mod] - prev))
            prev.add(mod)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import unittest
import sys
import os

import asn1crypto
from oscrypto import keys, backend

from .unittest_data import data_decorator, data
from ._unittest_compat import patch

patch()

if sys.version_info < (3,):
    byte_cls = str
else:
    byte_cls = bytes


tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')


_backend = backend()


if _backend == 'openssl':
    from oscrypto._openssl._libcrypto import libcrypto_legacy_support
    supports_legacy = libcrypto_legacy_support
else:
    supports_legacy = True


@data_decorator
class KeyTests(unittest.TestCase):

    @staticmethod
    def private_keys():
        return (
            (
                'keys/test-aes128.key',
                b'password123',
                'rsa',
                False,
            ),
            (
                'keys/test-aes256.key',
                b'password123',
                'rsa',
                False,
            ),
            (
                'keys/test-der.key',
                None,
                'rsa',
                False,
            ),
            (
                'keys/test-dsa-aes128.key',
                b'password123',
                'dsa',
                False,
            ),
            (
                'keys/test-dsa-der.key',
                None,
                'dsa',
                False,
            ),
            (
                'keys/test-dsa.key',
                None,
                'dsa',
                False,
            ),
            (
                'keys/test-ec-aes128.key',
                b'password123',
                'ec',
                False,
            ),
            (
                'keys/test-ec-der.key',
                None,
                'ec',
                False,
            ),
            (
                'keys/test-ec.key',
                None,
                'ec',
                False,
            ),
            (
                'keys/test-inter.key',
                None,
                'rsa',
                False,
            ),
            (
                'keys/test-pkcs8-aes128-der.key',
                b'password123',
                'rsa',
                False,
            ),
            (
                'keys/test-pkcs8-aes256.key',
                b'password123',
                'rsa',
                False,
            ),
            (
                'keys/test-pkcs8-blank-der.key',
                b'',
                'rsa',
                True,
            ),
            (
                'keys/test-pkcs8-blank-der.key',
                None,
                'rsa',
                True,
            ),
            (
                'keys/test-pkcs8-blank.key',
                b'',
                'rsa',
                True,
            ),
            (
                'keys/test-pkcs8-blank.key',
                None,
                'rsa',
                True,
            ),
            (
                'keys/test-pkcs8-der.key',
                None,
                'rsa',
                False,
            ),
            (
                'keys/test-pkcs8-des.key',
                b'password123',
                'rsa',
                True,
            ),
            (
                'keys/test-pkcs8-tripledes.key',
                b'password123',
                'rsa',
                False,
            ),
            (
                'keys/test-pkcs8.key',
                None,
                'rsa',
                False,
            ),
            (
                'keys/test-third-der.key',
                None,
                'rsa',
                False,
            ),
            (
                'keys/test-third.key',
                None,
                'rsa',
                False,
            ),
            (
                'keys/test-tripledes.key',
                b'password123',
                'rsa',
                False,
            ),
            (
                'keys/test.key',
                None,
                'rsa',
                False,
            ),
        )

    @data('private_keys')
    def parse_private(self, input_filename, password, algo, uses_legacy):
        def do_run():
            with open(os.path.join(fixtures_dir, input_filename), 'rb') as f:
                private_object = keys.parse_private(f.read(), password)

            self.assertEqual(algo, private_object['private_key_algorithm']['algorithm'].native)

            # Make sure we can parse the whole structure
            private_object.native

        if not supports_legacy and uses_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_parse_private_pem_leading_whitespace(self):
        with open(os.path.join(fixtures_dir, 'keys/test.key'), 'rb') as f:
            private_object = keys.parse_private(b'   \n' + f.read(), None)

        # Make sure we can parse the whole structure
        private_object.native

    @staticmethod
    def public_keys():
        return (
            (
                'keys/test-public-dsa-der.key',
                'dsa',
            ),
            (
                'keys/test-public-dsa.key',
                'dsa',
            ),
            (
                'keys/test-public-ec-der.key',
                'ec',
            ),
            (
                'keys/test-public-ec.key',
                'ec',
            ),
            (
                'keys/test-public-rsa-der.key',
                'rsa',
            ),
            (
                'keys/test-public-rsa.key',
                'rsa',
            ),
            (
                'keys/test-public-rsapublickey-der.key',
                'rsa',
            ),
            (
                'keys/test-public-rsapublickey.key',
                'rsa',
            ),
        )

    @data('public_keys')
    def parse_public(self, input_filename, algo):
        with open(os.path.join(fixtures_dir, input_filename), 'rb') as f:
            parsed = keys.parse_public(f.read())

        self.assertEqual(algo, parsed['algorithm']['algorithm'].native)

        # Make sure we can parse the whole structure
        parsed.native

    def test_parse_public_pem_leading_whitespace(self):
        with open(os.path.join(fixtures_dir, 'keys/test-public-rsa.key'), 'rb') as f:
            parsed = keys.parse_public(b'  \r\n' + f.read())

        # Make sure we can parse the whole structure
        parsed.native

    @staticmethod
    def certificates():
        return (
            (
                'keys/test-der.crt',
                'rsa'
            ),
            (
                'keys/test-dsa-der.crt',
                'dsa'
            ),
            (
                'keys/test-dsa.crt',
                'dsa'
            ),
            (
                'keys/test-ec-der.crt',
                'ec'
            ),
            (
                'keys/test-ec.crt',
                'ec'
            ),
            (
                'keys/test-inter-der.crt',
                'rsa'
            ),
            (
                'keys/test-inter.crt',
                'rsa'
            ),
            (
                'keys/test-third-der.crt',
                'rsa'
            ),
            (
                'keys/test-third.crt',
                'rsa'
            ),
            (
                'keys/test.crt',
                'rsa'
            ),
        )

    @data('certificates')
    def parse_certificate(self, input_filename, algo):
        with open(os.path.join(fixtures_dir, input_filename), 'rb') as f:
            parsed = keys.parse_certificate(f.read())

        self.assertEqual(algo, parsed['tbs_certificate']['subject_public_key_info']['algorithm']['algorithm'].native)
        self.assertEqual('Codex Non Sufficit LC', parsed['tbs_certificate']['subject'].native['organization_name'])

        # Make sure we can parse the whole structure
        parsed.native

    def test_parse_certificate_pem_leading_whitespace(self):
        with open(os.path.join(fixtures_dir, 'keys/test.crt'), 'rb') as f:
            parsed = keys.parse_certificate(b'\n' + f.read())

        # Make sure we can parse the whole structure
        parsed.native

    @staticmethod
    def pkcs12_files():
        return (
            (
                'aes128',
                'keys/test-aes128.p12',
                b'password123',
                True,
            ),
            (
                'aes256',
                'keys/test-aes256.p12',
                b'password123',
                True,
            ),
            (
                'rc2',
                'keys/test-rc2.p12',
                b'password123',
                True,
            ),
            (
                'tripledes_blank',
                'keys/test-tripledes-blank.p12',
                b'',
                True
            ),
            (
                'tripledes_blank_none',
                'keys/test-tripledes-blank.p12',
                None,
                True
            ),
            (
                'tripledes',
                'keys/test-tripledes.p12',
                b'password123',
                True
            ),
        )

    @data('pkcs12_files', True)
    def parse_pkcs12(self, input_filename, password, uses_legacy):
        def do_run():
            with open(os.path.join(fixtures_dir, input_filename), 'rb') as f:
                key_info, cert_info, extra_cert_infos = keys.parse_pkcs12(f.read(), password)

            with open(os.path.join(fixtures_dir, 'keys/test-pkcs8-der.key'), 'rb') as f:
                key_der = f.read()

            with open(os.path.join(fixtures_dir, 'keys/test-der.crt'), 'rb') as f:
                cert_der = f.read()

            self.assertEqual(key_der, key_info.dump())
            self.assertEqual(cert_der, cert_info.dump())
            self.assertEqual([], extra_cert_infos)

            # Make sure we can parse the DER
            key_info.native
            cert_info.native

        if not supports_legacy and uses_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_parse_pkcs12_dsa(self):
        def do_run():
            with open(os.path.join(fixtures_dir, 'keys/test-dsa.p12'), 'rb') as f:
                key_info, cert_info, extra_cert_infos = keys.parse_pkcs12(f.read(), b'password123')

            with open(os.path.join(fixtures_dir, 'keys/test-pkcs8-dsa-der.key'), 'rb') as f:
                key_der = f.read()

            with open(os.path.join(fixtures_dir, 'keys/test-dsa-der.crt'), 'rb') as f:
                cert_der = f.read()

            self.assertEqual(key_der, key_info.dump())
            self.assertEqual(cert_der, cert_info.dump())
            self.assertEqual([], extra_cert_infos)

            # Make sure we can parse the DER
            key_info.native
            cert_info.native

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_parse_pkcs12_chain(self):
        def do_run():
            with open(os.path.join(fixtures_dir, 'keys/test-third.p12'), 'rb') as f:
                key_info, cert_info, extra_cert_infos = keys.parse_pkcs12(f.read(), b'password123')

            with open(os.path.join(fixtures_dir, 'keys/test-third-der.key'), 'rb') as f:
                private_key = asn1crypto.keys.RSAPrivateKey.load(f.read())
                key_der = asn1crypto.keys.PrivateKeyInfo.wrap(private_key, 'rsa').dump()

            with open(os.path.join(fixtures_dir, 'keys/test-third-der.crt'), 'rb') as f:
                cert_der = f.read()

            with open(os.path.join(fixtures_dir, 'keys/test-inter-der.crt'), 'rb') as f:
                intermediate_cert_der = f.read()

            with open(os.path.join(fixtures_dir, 'keys/test-der.crt'), 'rb') as f:
                root_cert_der = f.read()

            self.assertEqual(key_der, key_info.dump())
            self.assertEqual(cert_der, cert_info.dump())
            self.assertEqual(
                sorted([intermediate_cert_der, root_cert_der]),
                sorted([info.dump() for info in extra_cert_infos])
            )

            # Make sure we can parse the DER
            key_info.native
            cert_info.native
            for info in extra_cert_infos:
                info.native

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()



# coding: utf-8

"""
This file was originally derived from
https://github.com/pypa/pip/blob/3e713708088aedb1cde32f3c94333d6e29aaf86e/src/pip/_internal/pep425tags.py

The following license covers that code:

Copyright (c) 2008-2018 The pip developers (see AUTHORS.txt file)

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from __future__ import unicode_literals, division, absolute_import, print_function

import sys
import os
import ctypes
import re
import platform

if sys.version_info >= (2, 7):
    import sysconfig

if sys.version_info < (3,):
    str_cls = unicode  # noqa
else:
    str_cls = str


def _pep425_implementation():
    """
    :return:
        A 2 character unicode string of the implementation - 'cp' for cpython
        or 'pp' for PyPy
    """

    return 'pp' if hasattr(sys, 'pypy_version_info') else 'cp'


def _pep425_version():
    """
    :return:
        A tuple of integers representing the Python version number
    """

    if hasattr(sys, 'pypy_version_info'):
        return (sys.version_info[0], sys.pypy_version_info.major,
                sys.pypy_version_info.minor)
    else:
        return (sys.version_info[0], sys.version_info[1])


def _pep425_supports_manylinux():
    """
    :return:
        A boolean indicating if the machine can use manylinux1 packages
    """

    try:
        import _manylinux
        return bool(_manylinux.manylinux1_compatible)
    except (ImportError, AttributeError):
        pass

    # Check for glibc 2.5
    try:
        proc = ctypes.CDLL(None)
        gnu_get_libc_version = proc.gnu_get_libc_version
        gnu_get_libc_version.restype = ctypes.c_char_p

        ver = gnu_get_libc_version()
        if not isinstance(ver, str_cls):
            ver = ver.decode('ascii')
        match = re.match(r'(\d+)\.(\d+)', ver)
        return match and match.group(1) == '2' and int(match.group(2)) >= 5

    except (AttributeError):
        return False


def _pep425_get_abi():
    """
    :return:
        A unicode string of the system abi. Will be something like: "cp27m",
        "cp33m", etc.
    """

    try:
        soabi = sysconfig.get_config_var('SOABI')
        if soabi:
            if soabi.startswith('cpython-'):
                return 'cp%s' % soabi.split('-')[1]
            return soabi.replace('.', '_').replace('-', '_')
    except (IOError, NameError):
        pass

    impl = _pep425_implementation()
    suffix = ''
    if impl == 'cp':
        suffix += 'm'
    if sys.maxunicode == 0x10ffff and sys.version_info < (3, 3):
        suffix += 'u'
    return '%s%s%s' % (impl, ''.join(map(str_cls, _pep425_version())), suffix)


def _pep425tags():
    """
    :return:
        A list of 3-element tuples with unicode strings or None:
         [0] implementation tag - cp33, pp27, cp26, py2, py2.py3
         [1] abi tag - cp26m, None
         [2] arch tag - linux_x86_64, macosx_10_10_x85_64, etc
    """

    tags = []

    versions = []
    version_info = _pep425_version()
    major = version_info[:-1]
    for minor in range(version_info[-1], -1, -1):
        versions.append(''.join(map(str, major + (minor,))))

    impl = _pep425_implementation()

    abis = []
    abi = _pep425_get_abi()
    if abi:
        abis.append(abi)
    abi3 = _pep425_implementation() == 'cp' and sys.version_info >= (3,)
    if abi3:
        abis.append('abi3')
    abis.append('none')

    if sys.platform == 'darwin':
        plat_ver = platform.mac_ver()
        ver_parts = plat_ver[0].split('.')
        minor = int(ver_parts[1])
        arch = plat_ver[2]
        if sys.maxsize == 2147483647:
            arch = 'i386'
        arches = []
        while minor > 5:
            arches.append('macosx_10_%s_%s' % (minor, arch))
            arches.append('macosx_10_%s_intel' % (minor,))
            arches.append('macosx_10_%s_universal' % (minor,))
            minor -= 1
    else:
        if sys.platform == 'win32':
            if 'amd64' in sys.version.lower():
                arches = ['win_amd64']
            arches = [sys.platform]
        elif hasattr(os, 'uname'):
            (plat, _, _, _, machine) = os.uname()
            plat = plat.lower().replace('/', '')
            machine.replace(' ', '_').replace('/', '_')
            if plat == 'linux' and sys.maxsize == 2147483647:
                machine = 'i686'
            arch = '%s_%s' % (plat, machine)
            if _pep425_supports_manylinux():
                arches = [arch.replace('linux', 'manylinux1'), arch]
            else:
                arches = [arch]

    for abi in abis:
        for arch in arches:
            tags.append(('%s%s' % (impl, versions[0]), abi, arch))

    if abi3:
        for version in versions[1:]:
            for arch in arches:
                tags.append(('%s%s' % (impl, version), 'abi3', arch))

    for arch in arches:
        tags.append(('py%s' % (versions[0][0]), 'none', arch))

    tags.append(('%s%s' % (impl, versions[0]), 'none', 'any'))
    tags.append(('%s%s' % (impl, versions[0][0]), 'none', 'any'))

    for i, version in enumerate(versions):
        tags.append(('py%s' % (version,), 'none', 'any'))
        if i == 0:
            tags.append(('py%s' % (version[0]), 'none', 'any'))

    tags.append(('py2.py3', 'none', 'any'))

    return tags



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os
import subprocess
import sys


run_args = [
    {
        'name': 'version',
        'kwarg': 'version',
    },
]


def _write_env(env, key, value):
    sys.stdout.write("%s: %s\n" % (key, value))
    sys.stdout.flush()
    if sys.version_info < (3,):
        env[key.encode('utf-8')] = value.encode('utf-8')
    else:
        env[key] = value


def run(version=None):
    """
    Installs a version of Python on Mac using pyenv

    :return:
        A bool - if Python was installed successfully
    """

    if sys.platform == 'win32':
        raise ValueError('pyenv-install is not designed for Windows')

    if version not in set(['2.6', '3.3']):
        raise ValueError('Invalid version: %r' % version)

    python_path = os.path.expanduser('~/.pyenv/versions/%s/bin' % version)
    if os.path.exists(os.path.join(python_path, 'python')):
        print(python_path)
        return True

    stdout = ""
    stderr = ""

    proc = subprocess.Popen(
        'command -v pyenv',
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE
    )
    proc.communicate()
    if proc.returncode != 0:
        proc = subprocess.Popen(
            ['brew', 'install', 'pyenv'],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        so, se = proc.communicate()
        stdout += so.decode('utf-8')
        stderr += se.decode('utf-8')
        if proc.returncode != 0:
            print(stdout)
            print(stderr, file=sys.stderr)
            return False

    pyenv_script = './%s' % version
    try:
        with open(pyenv_script, 'wb') as f:
            if version == '2.6':
                contents = '#require_gcc\n' \
                    'install_package "openssl-1.0.2k" "https://www.openssl.org/source/old/1.0.2/openssl-1.0.2k.tar.gz' \
                    '#6b3977c61f2aedf0f96367dcfb5c6e578cf37e7b8d913b4ecb6643c3cb88d8c0" mac_openssl\n' \
                    'install_package "readline-8.0" "https://ftpmirror.gnu.org/readline/readline-8.0.tar.gz' \
                    '#e339f51971478d369f8a053a330a190781acb9864cf4c541060f12078948e461" mac_readline' \
                    ' --if has_broken_mac_readline\n' \
                    'install_package "Python-2.6.9" "https://www.python.org/ftp/python/2.6.9/Python-2.6.9.tgz' \
                    '#7277b1285d8a82f374ef6ebaac85b003266f7939b3f2a24a3af52f9523ac94db" standard verify_py26'
            elif version == '3.3':
                contents = '#require_gcc\n' \
                    'install_package "openssl-1.0.2k" "https://www.openssl.org/source/old/1.0.2/openssl-1.0.2k.tar.gz' \
                    '#6b3977c61f2aedf0f96367dcfb5c6e578cf37e7b8d913b4ecb6643c3cb88d8c0" mac_openssl\n' \
                    'install_package "readline-8.0" "https://ftpmirror.gnu.org/readline/readline-8.0.tar.gz' \
                    '#e339f51971478d369f8a053a330a190781acb9864cf4c541060f12078948e461" mac_readline' \
                    ' --if has_broken_mac_readline\n' \
                    'install_package "Python-3.3.7" "https://www.python.org/ftp/python/3.3.7/Python-3.3.7.tar.xz' \
                    '#85f60c327501c36bc18c33370c14d472801e6af2f901dafbba056f61685429fe" standard verify_py33'
            f.write(contents.encode('utf-8'))

        args = ['pyenv', 'install', pyenv_script]
        stdin = None
        stdin_contents = None
        env = os.environ.copy()

        if version == '2.6':
            _write_env(env, 'PYTHON_CONFIGURE_OPTS', '--enable-ipv6')
            stdin = subprocess.PIPE
            stdin_contents = '--- configure  2021-08-05 20:17:26.000000000 -0400\n' \
                '+++ configure   2021-08-05 20:21:30.000000000 -0400\n' \
                '@@ -10300,17 +10300,8 @@\n' \
                ' rm -f core conftest.err conftest.$ac_objext \\\n' \
                '     conftest$ac_exeext conftest.$ac_ext\n' \
                ' \n' \
                '-if test "$buggygetaddrinfo" = "yes"; then\n' \
                '-\tif test "$ipv6" = "yes"; then\n' \
                '-\t\techo \'Fatal: You must get working getaddrinfo() function.\'\n' \
                '-\t\techo \'       or you can specify "--disable-ipv6"\'.\n' \
                '-\t\texit 1\n' \
                '-\tfi\n' \
                '-else\n' \
                '-\n' \
                ' $as_echo "#define HAVE_GETADDRINFO 1" >>confdefs.h\n' \
                ' \n' \
                '-fi\n' \
                ' for ac_func in getnameinfo\n' \
                ' do :\n' \
                '   ac_fn_c_check_func "$LINENO" "getnameinfo" "ac_cv_func_getnameinfo"'
            stdin_contents = stdin_contents.encode('ascii')
            args.append('--patch')

        proc = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=stdin,
            env=env
        )
        so, se = proc.communicate(stdin_contents)
        stdout += so.decode('utf-8')
        stderr += se.decode('utf-8')

        if proc.returncode != 0:
            print(stdout)
            print(stderr, file=sys.stderr)
            return False

    finally:
        if os.path.exists(pyenv_script):
            os.unlink(pyenv_script)

    print(python_path)
    return True



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os
import site
import sys

from . import build_root, requires_oscrypto
from ._import import _preload


deps_dir = os.path.join(build_root, 'modularcrypto-deps')
if os.path.exists(deps_dir):
    site.addsitedir(deps_dir)
    # In case any of the deps are installed system-wide
    sys.path.insert(0, deps_dir)

if sys.version_info[0:2] not in [(2, 6), (3, 2)]:
    from .lint import run as run_lint
else:
    run_lint = None

if sys.version_info[0:2] != (3, 2):
    from .coverage import run as run_coverage
    from .coverage import coverage
    run_tests = None

else:
    from .tests import run as run_tests
    run_coverage = None


def run():
    """
    Runs the linter and tests

    :return:
        A bool - if the linter and tests ran successfully
    """

    _preload(requires_oscrypto, True)

    if run_lint:
        print('')
        lint_result = run_lint()
    else:
        lint_result = True

    if run_coverage:
        print('\nRunning tests (via coverage.py %s)' % coverage.__version__)
        sys.stdout.flush()
        tests_result = run_coverage(ci=True)
    else:
        print('\nRunning tests')
        sys.stdout.flush()
        tests_result = run_tests(ci=True)
    sys.stdout.flush()

    return lint_result and tests_result



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os

from . import package_name, package_root

import flake8
if not hasattr(flake8, '__version_info__') or flake8.__version_info__ < (3,):
    from flake8.engine import get_style_guide
else:
    from flake8.api.legacy import get_style_guide


def run():
    """
    Runs flake8 lint

    :return:
        A bool - if flake8 did not find any errors
    """

    print('Running flake8 %s' % flake8.__version__)

    flake8_style = get_style_guide(config_file=os.path.join(package_root, 'tox.ini'))

    paths = []
    for _dir in [package_name, 'dev', 'tests']:
        for root, _, filenames in os.walk(_dir):
            for filename in filenames:
                if not filename.endswith('.py'):
                    continue
                paths.append(os.path.join(root, filename))
    report = flake8_style.check_files(paths)
    success = report.total_errors == 0
    if success:
        print('OK')
    return success



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import unittest
import sys

from oscrypto import symmetric, util, backend

from ._unittest_compat import patch
from .exception_context import assert_exception
from .unittest_data import data_decorator, data

patch()

if sys.version_info < (3,):
    byte_cls = str
else:
    byte_cls = bytes


_backend = backend()


if _backend == 'openssl':
    from oscrypto._openssl._libcrypto import libcrypto_legacy_support
    supports_legacy = libcrypto_legacy_support
else:
    supports_legacy = True


@data_decorator
class SymmetricTests(unittest.TestCase):

    @staticmethod
    def aes_key_lengths():
        return (
            ('128', 16),
            ('192', 24),
            ('256', 32),
        )

    @data('aes_key_lengths', True)
    def aes_cbc_no_padding_encrypt_decrypt(self, key_byte_len):
        key = util.rand_bytes(key_byte_len)
        data = b'This is the data'

        iv, ciphertext = symmetric.aes_cbc_no_padding_encrypt(key, data, None)
        self.assertNotEqual(data, ciphertext)
        self.assertEqual(byte_cls, type(ciphertext))

        plaintext = symmetric.aes_cbc_no_padding_decrypt(key, ciphertext, iv)
        self.assertEqual(data, plaintext)

    @data('aes_key_lengths', True)
    def aes_cbc_no_padding_encrypt_decrypt_two_block(self, key_byte_len):
        key = util.rand_bytes(key_byte_len)

        data = b'This is data to encrypt-32 bytes'

        iv, ciphertext = symmetric.aes_cbc_no_padding_encrypt(key, data, None)
        self.assertNotEqual(data, ciphertext)
        self.assertEqual(byte_cls, type(ciphertext))

        plaintext = symmetric.aes_cbc_no_padding_decrypt(key, ciphertext, iv)
        self.assertEqual(data, plaintext)

    @data('aes_key_lengths', True)
    def aes_cbc_no_padding_wrong_length(self, key_byte_len):
        key = util.rand_bytes(key_byte_len)

        with assert_exception(self, ValueError, r'data must be a multiple of 16 bytes long - is 31'):
            data = b'31 bytes of data to encrypt now'
            iv, ciphertext = symmetric.aes_cbc_no_padding_encrypt(key, data, None)

        with assert_exception(self, ValueError, r'data must be a multiple of 16 bytes long - is 33'):
            data = b'Thirty three bytes to encrypt now'
            iv, ciphertext = symmetric.aes_cbc_no_padding_encrypt(key, data, None)

        with assert_exception(self, ValueError, r'data must be a multiple of 16 bytes long - is 15'):
            data = b'Fifteen bytes!!'
            iv, ciphertext = symmetric.aes_cbc_no_padding_encrypt(key, data, None)

        with assert_exception(self, ValueError, r'data must be a multiple of 16 bytes long - is 24'):
            data = b'Twenty four bytes long!!'
            iv, ciphertext = symmetric.aes_cbc_no_padding_encrypt(key, data, None)

    @data('aes_key_lengths', True)
    def aes_encrypt_decrypt(self, key_byte_len):
        key = util.rand_bytes(key_byte_len)
        data = b'This is data to encrypt'

        iv, ciphertext = symmetric.aes_cbc_pkcs7_encrypt(key, data, None)
        self.assertNotEqual(data, ciphertext)
        self.assertEqual(byte_cls, type(ciphertext))

        plaintext = symmetric.aes_cbc_pkcs7_decrypt(key, ciphertext, iv)
        self.assertEqual(data, plaintext)

    def test_rc4_40_encrypt_decrypt(self):
        def do_run():
            key = util.rand_bytes(5)
            data = b'This is data to encrypt'

            ciphertext = symmetric.rc4_encrypt(key, data)
            self.assertNotEqual(data, ciphertext)
            self.assertEqual(byte_cls, type(ciphertext))

            plaintext = symmetric.rc4_decrypt(key, ciphertext)
            self.assertEqual(data, plaintext)

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_rc4_128_encrypt_decrypt(self):
        def do_run():
            key = util.rand_bytes(16)
            data = b'This is data to encrypt'

            ciphertext = symmetric.rc4_encrypt(key, data)
            self.assertNotEqual(data, ciphertext)
            self.assertEqual(byte_cls, type(ciphertext))

            plaintext = symmetric.rc4_decrypt(key, ciphertext)
            self.assertEqual(data, plaintext)

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_rc2_64_encrypt_decrypt(self):
        def do_run():
            key = util.rand_bytes(8)
            data = b'This is data to encrypt'

            iv, ciphertext = symmetric.rc2_cbc_pkcs5_encrypt(key, data, None)
            self.assertNotEqual(data, ciphertext)
            self.assertEqual(byte_cls, type(ciphertext))

            plaintext = symmetric.rc2_cbc_pkcs5_decrypt(key, ciphertext, iv)
            self.assertEqual(data, plaintext)

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_rc2_40_encrypt_decrypt(self):
        def do_run():
            key = util.rand_bytes(5)
            data = b'This is data to encrypt'

            iv, ciphertext = symmetric.rc2_cbc_pkcs5_encrypt(key, data, None)
            self.assertNotEqual(data, ciphertext)
            self.assertEqual(byte_cls, type(ciphertext))

            plaintext = symmetric.rc2_cbc_pkcs5_decrypt(key, ciphertext, iv)
            self.assertEqual(data, plaintext)

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_des_encrypt_decrypt(self):
        def do_run():
            key = util.rand_bytes(8)
            data = b'This is data to encrypt'

            iv, ciphertext = symmetric.des_cbc_pkcs5_encrypt(key, data, None)
            self.assertNotEqual(data, ciphertext)
            self.assertEqual(byte_cls, type(ciphertext))

            plaintext = symmetric.des_cbc_pkcs5_decrypt(key, ciphertext, iv)
            self.assertEqual(data, plaintext)

        if not supports_legacy:
            with self.assertRaises(EnvironmentError):
                do_run()
        else:
            do_run()

    def test_3des_2k_encrypt_decrypt(self):
        key = util.rand_bytes(16)
        data = b'This is data to encrypt'

        iv, ciphertext = symmetric.tripledes_cbc_pkcs5_encrypt(key, data, None)
        self.assertNotEqual(data, ciphertext)
        self.assertEqual(byte_cls, type(ciphertext))

        plaintext = symmetric.tripledes_cbc_pkcs5_decrypt(key, ciphertext, iv)
        self.assertEqual(data, plaintext)

    def test_3des_3k_encrypt_decrypt(self):
        key = util.rand_bytes(24)
        data = b'This is data to encrypt'

        iv, ciphertext = symmetric.tripledes_cbc_pkcs5_encrypt(key, data, None)
        self.assertNotEqual(data, ciphertext)
        self.assertEqual(byte_cls, type(ciphertext))

        plaintext = symmetric.tripledes_cbc_pkcs5_decrypt(key, ciphertext, iv)
        self.assertEqual(data, plaintext)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import hashlib
import os
import unittest
import sys

from oscrypto import trust_list
from asn1crypto import x509, pem

from ._unittest_compat import patch

patch()

if sys.version_info < (3,):
    str_cls = unicode  # noqa
    byte_cls = str
else:
    str_cls = str
    byte_cls = bytes


tests_root = os.path.dirname(__file__)
fixtures_dir = os.path.join(tests_root, 'fixtures')

digicert_ca_path = os.path.join(fixtures_dir, 'digicert_ca.crt')


class TrustListTests(unittest.TestCase):

    def test_get_list(self):
        trust_list.clear_cache()

        certs = trust_list.get_list()
        self.assertIsInstance(certs, list)
        self.assertLess(10, len(certs))
        for cert, trust_oids, reject_oids in certs:
            self.assertIsInstance(cert, x509.Certificate)
            self.assertIsInstance(trust_oids, set)
            self.assertIsInstance(reject_oids, set)
            cert.native

    def test_get_list_callback(self):
        trust_list.clear_cache()

        lambda_data = {'calls': 0, 'reasons': 0, 'certs': {}}

        def cb(cert, reason):
            if reason is not None:
                self.assertIsInstance(reason, str_cls)
                lambda_data['reasons'] += 1
            self.assertIsInstance(cert, x509.Certificate)
            sha1 = hashlib.sha1(cert.dump()).digest()
            message = None
            if sha1 in lambda_data['certs']:
                message = 'Certificate (%s) already passed to callback' % cert.subject.human_friendly
            self.assertNotIn(sha1, lambda_data['certs'], message)
            lambda_data['certs'][sha1] = True
            lambda_data['calls'] += 1

        certs = trust_list.get_list(cert_callback=cb)
        self.assertIsInstance(certs, list)
        self.assertLess(10, len(certs))
        self.assertLessEqual(len(certs), lambda_data['calls'])
        self.assertEqual(lambda_data['calls'] - len(certs), lambda_data['reasons'])
        for cert, trust_oids, reject_oids in certs:
            self.assertIsInstance(cert, x509.Certificate)
            self.assertIsInstance(trust_oids, set)
            self.assertIsInstance(reject_oids, set)
            cert.native

    def test_get_list_mutate(self):
        trust_list.clear_cache()

        certs = trust_list.get_list()
        certs2 = trust_list.get_list()

        with open(digicert_ca_path, 'rb') as f:
            _, _, digicert_ca_bytes = pem.unarmor(f.read())
            digicert_ca_cert = x509.Certificate.load(digicert_ca_bytes)
        certs.append(digicert_ca_cert)

        self.assertNotEqual(certs2, certs)

    def test_get_path(self):
        trust_list.clear_cache()

        certs = trust_list.get_path()
        with open(certs, 'rb') as f:
            cert_data = f.read()
            self.assertEqual(True, pem.detect(cert_data))
            self.assertLess(10240, len(cert_data))



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function


import socket
import select
import threading


_sockets = {}
_socket_lock = threading.Lock()


def proxy(src, dst, callback=None):
    timeout = 10
    try:
        read_ready, _, _ = select.select([src], [], [], timeout)
        while len(read_ready):
            if callback:
                callback(src, dst)
            else:
                dst.send(src.recv(8192))
            read_ready, _, _ = select.select([src], [], [], timeout)
    except (socket.error, select.error, OSError, ValueError):
        pass
    try:
        src.shutdown(socket.SHUT_RDWR)
    except (socket.error, OSError, ValueError):
        pass
    src.close()
    try:
        dst.shutdown(socket.SHUT_RDWR)
    except (socket.error, OSError, ValueError):
        pass
    dst.close()


def listen(server, ip, port, send_callback, recv_callback):
    lsock, laddr = server.accept()
    rsock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    rsock.connect((ip, port))
    with _socket_lock:
        _sockets[threading.current_thread().ident] = {
            'lsock': lsock,
            'rsock': rsock
        }
    t1 = threading.Thread(target=proxy, args=(rsock, lsock, recv_callback))
    t2 = threading.Thread(target=proxy, args=(lsock, rsock, send_callback))
    t1.start()
    t2.start()


def make_socket_proxy(ip, port, send_callback=None, recv_callback=None):
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(('', 8080))
    server.listen(1)
    t = threading.Thread(
        target=listen,
        args=(server, ip, port, send_callback, recv_callback)
    )
    t.start()
    sock = socket.create_connection(('localhost', 8080))
    sock.settimeout(1)
    t.join()
    with _socket_lock:
        data = _sockets[t.ident]
        return (sock, data['lsock'], data['rsock'], server)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys
import re
import os

from oscrypto import tls, errors as oscrypto_errors, version
from asn1crypto.util import OrderedDict

if sys.version_info < (3,):
    from urlparse import urlparse
    str_cls = unicode  # noqa
else:
    from urllib.parse import urlparse
    str_cls = str


class HttpsClientException(Exception):
    pass


class HttpsClientError(HttpsClientException):
    pass


class HttpsClient():

    def __init__(self, keep_alive=True, ignore_close=False):
        self.socket = None
        self.timeout = None
        self.url_info = None
        self.keep_alive = keep_alive
        self.ignore_close = ignore_close

    def close(self):
        """
        Closes any open connection
        """

        if not self.socket:
            return
        self.socket.close()
        self.socket = None

    def download(self, url, timeout):
        """
        Downloads a URL and returns the contents

        :param url:
            The URL to download

        :param timeout:
            The int number of seconds to set the timeout to

        :return:
            The string contents of the URL
        """

        self.setup_connection(url, timeout)
        tries = 0

        while tries < 2:
            tries += 1
            try:
                self.ensure_connected()

                req_headers = OrderedDict()
                req_headers['Host'] = self.url_info[0]
                if self.url_info[1] != 443:
                    req_headers['Host'] += ':%d' % self.url_info[1]
                req_headers['Connection'] = 'Keep-Alive' if self.keep_alive else 'Close'
                req_headers["User-Agent"] = 'oscrypto %s TLS HTTP Client' % version.__version__

                request = 'GET '
                url_info = urlparse(url)
                path = '/' if not url_info.path else url_info.path
                if url_info.query:
                    path += '?' + url_info.query
                request += path + ' HTTP/1.1'
                self.write_request(request, req_headers)

                response = self.read_headers()
                if not response:
                    self.close()
                    continue

                v, code, message, resp_headers = response
                data = self.read_body(code, resp_headers, timeout)

                if code == 301:
                    location = resp_headers.get('location')
                    if not isinstance(location, str_cls):
                        raise HttpsClientError('Missing or duplicate Location HTTP header')
                    if not re.match(r'https?://', location):
                        if not location.startswith('/'):
                            location = os.path.dirname(url_info.path) + location
                        location = url_info.scheme + '://' + url_info.netloc + location
                    return self.download(location, timeout)

                if code != 200:
                    raise HttpsClientError('HTTP error %s downloading %s.' % (code, url))

                else:
                    return data

            except (oscrypto_errors.TLSGracefulDisconnectError):
                self.close()
                continue

    def setup_connection(self, url, timeout):
        """
        :param url:
            The URL to download

        :param timeout:
            The int number of seconds to set the timeout to

        :return:
            A boolean indicating if the connection was reused
        """

        url_info = urlparse(url)
        if url_info.scheme == 'http':
            raise HttpsClientException('Can not connect to a non-TLS server')
        hostname = url_info.hostname
        port = url_info.port
        if not port:
            port = 443

        if self.socket and self.url_info != (hostname, port):
            self.close()

        self.timeout = timeout
        self.url_info = (hostname, port)

        return self.ensure_connected()

    def ensure_connected(self):
        """
        Make sure a valid tls.TLSSocket() is open to the server

        :return:
            A boolean indicating if the connection was reused
        """

        if self.socket:
            return True

        host, port = self.url_info
        session = tls.TLSSession()
        self.socket = tls.TLSSocket(host, port, timeout=self.timeout, session=session)
        return False

    def write_request(self, request, headers):
        """
        :param request:
            A unicode string of the first line of the HTTP request

        :param headers:
            An OrderedDict of the request headers
        """

        lines = [request]
        for header, value in headers.items():
            lines.append('%s: %s' % (header, value))
        lines.extend(['', ''])

        request = '\r\n'.join(lines).encode('iso-8859-1')
        self.socket.write(request)

    def read_headers(self):
        """
        Reads the HTTP response headers from the socket

        :return:
            On error, None, otherwise a 4-element tuple:
              0: A 2-element tuple of integers representing the HTTP version
              1: An integer representing the HTTP response code
              2: A unicode string of the HTTP response code name
              3: An OrderedDict of HTTP headers with lowercase unicode key and unicode values
        """

        version = None
        code = None
        text = None
        headers = OrderedDict()

        data = self.socket.read_until(b'\r\n\r\n')
        string = data.decode('iso-8859-1')
        first = False
        for line in string.split('\r\n'):
            line = line.strip()
            if first is False:
                if line == '':
                    continue
                match = re.match(r'^HTTP/(1\.[01]) +(\d+) +(.*)$', line)
                if not match:
                    return None
                version = tuple(map(int, match.group(1).split('.')))
                code = int(match.group(2))
                text = match.group(3)
                first = True
            else:
                if not len(line):
                    continue
                parts = line.split(':', 1)
                if len(parts) == 2:
                    name = parts[0].strip().lower()
                    value = parts[1].strip()
                    if name in headers:
                        if isinstance(headers[name], tuple):
                            headers[name] = headers[name] + (value,)
                        else:
                            headers[name] = (headers[name], value)
                    else:
                        headers[name] = value

        return (version, code, text, headers)

    def parse_content_length(self, headers):
        """
        Returns the content-length from a dict of headers

        :return:
            An integer of the content length
        """

        content_length = headers.get('content-length')
        if isinstance(content_length, str_cls) and len(content_length) > 0:
            content_length = int(content_length)
        return content_length

    def read_body(self, code, resp_headers, timeout):
        """

        """

        data = b''
        transfer_encoding = resp_headers.get('transfer-encoding')
        if transfer_encoding and transfer_encoding.lower() == 'chunked':
            while True:
                line = self.socket.read_until(b'\r\n').decode('iso-8859-1').rstrip()
                if re.match(r'^[a-fA-F0-9]+$', line):
                    chunk_length = int(line, 16)
                    if chunk_length == 0:
                        break
                    data += self.socket.read_exactly(chunk_length)
                    if self.socket.read_exactly(2) != b'\r\n':
                        raise HttpsClientException('Unable to parse chunk newline')
                else:
                    self.close()
                    raise HttpsClientException('Unable to parse chunk length')
        else:
            content_length = self.parse_content_length(resp_headers)
            if content_length is not None:
                if content_length > 0:
                    data = self.socket.read_exactly(content_length)
            elif code == 304 or code == 204 or (code >= 100 and code < 200):
                # Per https://tools.ietf.org/html/rfc7230#section-3.3.3 these have no body
                pass
            else:
                # This should only happen if the server is going to close the connection
                while self.socket.select_read(timeout=timeout):
                    data += self.socket.read(8192)
                self.close()

        if not self.ignore_close and resp_headers.get('connection', '').lower() == 'close':
            self.close()

        return data



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import imp
import os
import tarfile
import zipfile

import setuptools.sandbox

from . import package_root, package_name, has_tests_package


def _list_zip(filename):
    """
    Prints all of the files in a .zip file
    """

    zf = zipfile.ZipFile(filename, 'r')
    for name in zf.namelist():
        print('     %s' % name)


def _list_tgz(filename):
    """
    Prints all of the files in a .tar.gz file
    """

    tf = tarfile.open(filename, 'r:gz')
    for name in tf.getnames():
        print('     %s' % name)


def run():
    """
    Creates a sdist .tar.gz and a bdist_wheel --univeral .whl

    :return:
        A bool - if the packaging process was successful
    """

    setup = os.path.join(package_root, 'setup.py')
    tests_root = os.path.join(package_root, 'tests')
    tests_setup = os.path.join(tests_root, 'setup.py')

    # Trying to call setuptools.sandbox.run_setup(setup, ['--version'])
    # resulted in a segfault, so we do this instead
    module_info = imp.find_module('version', [os.path.join(package_root, package_name)])
    version_mod = imp.load_module('%s.version' % package_name, *module_info)

    pkg_name_info = (package_name, version_mod.__version__)
    print('Building %s-%s' % pkg_name_info)

    sdist = '%s-%s.tar.gz' % pkg_name_info
    whl = '%s-%s-py2.py3-none-any.whl' % pkg_name_info
    setuptools.sandbox.run_setup(setup, ['-q', 'sdist'])
    print(' - created %s' % sdist)
    _list_tgz(os.path.join(package_root, 'dist', sdist))
    setuptools.sandbox.run_setup(setup, ['-q', 'bdist_wheel', '--universal'])
    print(' - created %s' % whl)
    _list_zip(os.path.join(package_root, 'dist', whl))
    setuptools.sandbox.run_setup(setup, ['-q', 'clean'])

    if has_tests_package:
        print('Building %s_tests-%s' % (package_name, version_mod.__version__))

        tests_sdist = '%s_tests-%s.tar.gz' % pkg_name_info
        tests_whl = '%s_tests-%s-py2.py3-none-any.whl' % pkg_name_info
        setuptools.sandbox.run_setup(tests_setup, ['-q', 'sdist'])
        print(' - created %s' % tests_sdist)
        _list_tgz(os.path.join(tests_root, 'dist', tests_sdist))
        setuptools.sandbox.run_setup(tests_setup, ['-q', 'bdist_wheel', '--universal'])
        print(' - created %s' % tests_whl)
        _list_zip(os.path.join(tests_root, 'dist', tests_whl))
        setuptools.sandbox.run_setup(tests_setup, ['-q', 'clean'])

        dist_dir = os.path.join(package_root, 'dist')
        tests_dist_dir = os.path.join(tests_root, 'dist')
        os.rename(
            os.path.join(tests_dist_dir, tests_sdist),
            os.path.join(dist_dir, tests_sdist)
        )
        os.rename(
            os.path.join(tests_dist_dir, tests_whl),
            os.path.join(dist_dir, tests_whl)
        )
        os.rmdir(tests_dist_dir)

    return True



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import cgi
import codecs
import coverage
import imp
import json
import os
import unittest
import re
import sys
import tempfile
import time
import platform as _plat
import subprocess
from fnmatch import fnmatch

from . import package_name, package_root, other_packages

if sys.version_info < (3,):
    str_cls = unicode  # noqa
    from urllib2 import URLError
    from urllib import urlencode
    from io import open
else:
    str_cls = str
    from urllib.error import URLError
    from urllib.parse import urlencode

if sys.version_info < (3, 7):
    Pattern = re._pattern_type
else:
    Pattern = re.Pattern


def run(ci=False):
    """
    Runs the tests while measuring coverage

    :param ci:
        If coverage is being run in a CI environment - this triggers trying to
        run the tests for the rest of modularcrypto and uploading coverage data

    :return:
        A bool - if the tests ran successfully
    """

    xml_report_path = os.path.join(package_root, 'coverage.xml')
    if os.path.exists(xml_report_path):
        os.unlink(xml_report_path)

    cov = coverage.Coverage(include='%s/*.py' % package_name)
    cov.start()

    from .tests import run as run_tests
    result = run_tests(ci=ci)
    print()

    if ci:
        suite = unittest.TestSuite()
        loader = unittest.TestLoader()
        for other_package in other_packages:
            for test_class in _load_package_tests(other_package):
                suite.addTest(loader.loadTestsFromTestCase(test_class))

        if suite.countTestCases() > 0:
            print('Running tests from other modularcrypto packages')
            sys.stdout.flush()
            runner_result = unittest.TextTestRunner(stream=sys.stdout, verbosity=1).run(suite)
            result = runner_result.wasSuccessful() and result
            print()
            sys.stdout.flush()

    cov.stop()
    cov.save()

    cov.report(show_missing=False)
    print()
    sys.stdout.flush()
    if ci:
        cov.xml_report()

    if ci and result and os.path.exists(xml_report_path):
        _codecov_submit()
        print()

    return result


def _load_package_tests(name):
    """
    Load the test classes from another modularcrypto package

    :param name:
        A unicode string of the other package name

    :return:
        A list of unittest.TestCase classes of the tests for the package
    """

    package_dir = os.path.join('..', name)
    if not os.path.exists(package_dir):
        return []

    tests_module_info = imp.find_module('tests', [package_dir])
    tests_module = imp.load_module('%s.tests' % name, *tests_module_info)
    return tests_module.test_classes()


def _env_info():
    """
    :return:
        A two-element tuple of unicode strings. The first is the name of the
        environment, the second the root of the repo. The environment name
        will be one of: "ci-travis", "ci-circle", "ci-appveyor",
        "ci-github-actions", "local"
    """

    if os.getenv('CI') == 'true' and os.getenv('TRAVIS') == 'true':
        return ('ci-travis', os.getenv('TRAVIS_BUILD_DIR'))

    if os.getenv('CI') == 'True' and os.getenv('APPVEYOR') == 'True':
        return ('ci-appveyor', os.getenv('APPVEYOR_BUILD_FOLDER'))

    if os.getenv('CI') == 'true' and os.getenv('CIRCLECI') == 'true':
        return ('ci-circle', os.getcwdu() if sys.version_info < (3,) else os.getcwd())

    if os.getenv('GITHUB_ACTIONS') == 'true':
        return ('ci-github-actions', os.getenv('GITHUB_WORKSPACE'))

    return ('local', package_root)


def _codecov_submit():
    env_name, root = _env_info()

    try:
        with open(os.path.join(root, 'dev/codecov.json'), 'rb') as f:
            json_data = json.loads(f.read().decode('utf-8'))
    except (OSError, ValueError, UnicodeDecodeError, KeyError):
        print('error reading codecov.json')
        return

    if json_data.get('disabled'):
        return

    if env_name == 'ci-travis':
        # http://docs.travis-ci.com/user/environment-variables/#Default-Environment-Variables
        build_url = 'https://travis-ci.org/%s/jobs/%s' % (os.getenv('TRAVIS_REPO_SLUG'), os.getenv('TRAVIS_JOB_ID'))
        query = {
            'service': 'travis',
            'branch': os.getenv('TRAVIS_BRANCH'),
            'build': os.getenv('TRAVIS_JOB_NUMBER'),
            'pr': os.getenv('TRAVIS_PULL_REQUEST'),
            'job': os.getenv('TRAVIS_JOB_ID'),
            'tag': os.getenv('TRAVIS_TAG'),
            'slug': os.getenv('TRAVIS_REPO_SLUG'),
            'commit': os.getenv('TRAVIS_COMMIT'),
            'build_url': build_url,
        }

    elif env_name == 'ci-appveyor':
        # http://www.appveyor.com/docs/environment-variables
        build_url = 'https://ci.appveyor.com/project/%s/build/%s' % (
            os.getenv('APPVEYOR_REPO_NAME'),
            os.getenv('APPVEYOR_BUILD_VERSION')
        )
        query = {
            'service': "appveyor",
            'branch': os.getenv('APPVEYOR_REPO_BRANCH'),
            'build': os.getenv('APPVEYOR_JOB_ID'),
            'pr': os.getenv('APPVEYOR_PULL_REQUEST_NUMBER'),
            'job': '/'.join((
                os.getenv('APPVEYOR_ACCOUNT_NAME'),
                os.getenv('APPVEYOR_PROJECT_SLUG'),
                os.getenv('APPVEYOR_BUILD_VERSION')
            )),
            'tag': os.getenv('APPVEYOR_REPO_TAG_NAME'),
            'slug': os.getenv('APPVEYOR_REPO_NAME'),
            'commit': os.getenv('APPVEYOR_REPO_COMMIT'),
            'build_url': build_url,
        }

    elif env_name == 'ci-circle':
        # https://circleci.com/docs/environment-variables
        query = {
            'service': 'circleci',
            'branch': os.getenv('CIRCLE_BRANCH'),
            'build': os.getenv('CIRCLE_BUILD_NUM'),
            'pr': os.getenv('CIRCLE_PR_NUMBER'),
            'job': os.getenv('CIRCLE_BUILD_NUM') + "." + os.getenv('CIRCLE_NODE_INDEX'),
            'tag': os.getenv('CIRCLE_TAG'),
            'slug': os.getenv('CIRCLE_PROJECT_USERNAME') + "/" + os.getenv('CIRCLE_PROJECT_REPONAME'),
            'commit': os.getenv('CIRCLE_SHA1'),
            'build_url': os.getenv('CIRCLE_BUILD_URL'),
        }

    elif env_name == 'ci-github-actions':
        branch = ''
        tag = ''
        ref = os.getenv('GITHUB_REF', '')
        if ref.startswith('refs/tags/'):
            tag = ref[10:]
        elif ref.startswith('refs/heads/'):
            branch = ref[11:]

        impl = _plat.python_implementation()
        major, minor = _plat.python_version_tuple()[0:2]
        build_name = '%s %s %s.%s' % (_platform_name(), impl, major, minor)

        query = {
            'service': 'custom',
            'token': json_data['token'],
            'branch': branch,
            'tag': tag,
            'slug': os.getenv('GITHUB_REPOSITORY'),
            'commit': os.getenv('GITHUB_SHA'),
            'build_url': 'https://github.com/wbond/oscrypto/commit/%s/checks' % os.getenv('GITHUB_SHA'),
            'name': 'GitHub Actions %s on %s' % (build_name, os.getenv('RUNNER_OS'))
        }

    else:
        if not os.path.exists(os.path.join(root, '.git')):
            print('git repository not found, not submitting coverage data')
            return
        git_status = _git_command(['status', '--porcelain'], root)
        if git_status != '':
            print('git repository has uncommitted changes, not submitting coverage data')
            return

        branch = _git_command(['rev-parse', '--abbrev-ref', 'HEAD'], root)
        commit = _git_command(['rev-parse', '--verify', 'HEAD'], root)
        tag = _git_command(['name-rev', '--tags', '--name-only', commit], root)
        impl = _plat.python_implementation()
        major, minor = _plat.python_version_tuple()[0:2]
        build_name = '%s %s %s.%s' % (_platform_name(), impl, major, minor)
        query = {
            'branch': branch,
            'commit': commit,
            'slug': json_data['slug'],
            'token': json_data['token'],
            'build': build_name,
        }
        if tag != 'undefined':
            query['tag'] = tag

    payload = 'PLATFORM=%s\n' % _platform_name()
    payload += 'PYTHON_VERSION=%s %s\n' % (_plat.python_version(), _plat.python_implementation())
    if 'oscrypto' in sys.modules:
        payload += 'OSCRYPTO_BACKEND=%s\n' % sys.modules['oscrypto'].backend()
    payload += '<<<<<< ENV\n'

    for path in _list_files(root):
        payload += path + '\n'
    payload += '<<<<<< network\n'

    payload += '# path=coverage.xml\n'
    with open(os.path.join(root, 'coverage.xml'), 'r', encoding='utf-8') as f:
        payload += f.read() + '\n'
    payload += '<<<<<< EOF\n'

    url = 'https://codecov.io/upload/v4'
    headers = {
        'Accept': 'text/plain'
    }
    filtered_query = {}
    for key in query:
        value = query[key]
        if value == '' or value is None:
            continue
        filtered_query[key] = value

    print('Submitting coverage info to codecov.io')
    info = _do_request(
        'POST',
        url,
        headers,
        query_params=filtered_query
    )

    encoding = info[1] or 'utf-8'
    text = info[2].decode(encoding).strip()
    parts = text.split()
    upload_url = parts[1]

    headers = {
        'Content-Type': 'text/plain',
        'x-amz-acl': 'public-read',
        'x-amz-storage-class': 'REDUCED_REDUNDANCY'
    }

    print('Uploading coverage data to codecov.io S3 bucket')
    _do_request(
        'PUT',
        upload_url,
        headers,
        data=payload.encode('utf-8')
    )


def _git_command(params, cwd):
    """
    Executes a git command, returning the output

    :param params:
        A list of the parameters to pass to git

    :param cwd:
        The working directory to execute git in

    :return:
        A 2-element tuple of (stdout, stderr)
    """

    proc = subprocess.Popen(
        ['git'] + params,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=cwd
    )
    stdout, stderr = proc.communicate()
    code = proc.wait()
    if code != 0:
        e = OSError('git exit code was non-zero')
        e.stdout = stdout
        raise e
    return stdout.decode('utf-8').strip()


def _parse_env_var_file(data):
    """
    Parses a basic VAR="value data" file contents into a dict

    :param data:
        A unicode string of the file data

    :return:
        A dict of parsed name/value data
    """

    output = {}
    for line in data.splitlines():
        line = line.strip()
        if not line or '=' not in line:
            continue
        parts = line.split('=')
        if len(parts) != 2:
            continue
        name = parts[0]
        value = parts[1]
        if len(value) > 1:
            if value[0] == '"' and value[-1] == '"':
                value = value[1:-1]
        output[name] = value
    return output


def _platform_name():
    """
    Returns information about the current operating system and version

    :return:
        A unicode string containing the OS name and version
    """

    if sys.platform == 'darwin':
        version = _plat.mac_ver()[0]
        _plat_ver_info = tuple(map(int, version.split('.')))
        if _plat_ver_info < (10, 12):
            name = 'OS X'
        else:
            name = 'macOS'
        return '%s %s' % (name, version)

    elif sys.platform == 'win32':
        _win_ver = sys.getwindowsversion()
        _plat_ver_info = (_win_ver[0], _win_ver[1])
        return 'Windows %s' % _plat.win32_ver()[0]

    elif sys.platform in ['linux', 'linux2']:
        if os.path.exists('/etc/os-release'):
            with open('/etc/os-release', 'r', encoding='utf-8') as f:
                pairs = _parse_env_var_file(f.read())
                if 'NAME' in pairs and 'VERSION_ID' in pairs:
                    return '%s %s' % (pairs['NAME'], pairs['VERSION_ID'])
                    version = pairs['VERSION_ID']
                elif 'PRETTY_NAME' in pairs:
                    return pairs['PRETTY_NAME']
                elif 'NAME' in pairs:
                    return pairs['NAME']
                else:
                    raise ValueError('No suitable version info found in /etc/os-release')
        elif os.path.exists('/etc/lsb-release'):
            with open('/etc/lsb-release', 'r', encoding='utf-8') as f:
                pairs = _parse_env_var_file(f.read())
                if 'DISTRIB_DESCRIPTION' in pairs:
                    return pairs['DISTRIB_DESCRIPTION']
                else:
                    raise ValueError('No suitable version info found in /etc/lsb-release')
        else:
            return 'Linux'

    else:
        return '%s %s' % (_plat.system(), _plat.release())


def _list_files(root):
    """
    Lists all of the files in a directory, taking into account any .gitignore
    file that is present

    :param root:
        A unicode filesystem path

    :return:
        A list of unicode strings, containing paths of all files not ignored
        by .gitignore with root, using relative paths
    """

    dir_patterns, file_patterns = _gitignore(root)
    paths = []
    prefix = os.path.abspath(root) + os.sep
    for base, dirs, files in os.walk(root):
        for d in dirs:
            for dir_pattern in dir_patterns:
                if fnmatch(d, dir_pattern):
                    dirs.remove(d)
                    break
        for f in files:
            skip = False
            for file_pattern in file_patterns:
                if fnmatch(f, file_pattern):
                    skip = True
                    break
            if skip:
                continue
            full_path = os.path.join(base, f)
            if full_path[:len(prefix)] == prefix:
                full_path = full_path[len(prefix):]
            paths.append(full_path)
    return sorted(paths)


def _gitignore(root):
    """
    Parses a .gitignore file and returns patterns to match dirs and files.
    Only basic gitignore patterns are supported. Pattern negation, ** wildcards
    and anchored patterns are not currently implemented.

    :param root:
        A unicode string of the path to the git repository

    :return:
        A 2-element tuple:
         - 0: a list of unicode strings to match against dirs
         - 1: a list of unicode strings to match against dirs and files
    """

    gitignore_path = os.path.join(root, '.gitignore')

    dir_patterns = ['.git']
    file_patterns = []

    if not os.path.exists(gitignore_path):
        return (dir_patterns, file_patterns)

    with open(gitignore_path, 'r', encoding='utf-8') as f:
        for line in f.readlines():
            line = line.strip()
            if not line:
                continue
            if line.startswith('#'):
                continue
            if '**' in line:
                raise NotImplementedError('gitignore ** wildcards are not implemented')
            if line.startswith('!'):
                raise NotImplementedError('gitignore pattern negation is not implemented')
            if line.startswith('/'):
                raise NotImplementedError('gitignore anchored patterns are not implemented')
            if line.startswith('\\#'):
                line = '#' + line[2:]
            if line.startswith('\\!'):
                line = '!' + line[2:]
            if line.endswith('/'):
                dir_patterns.append(line[:-1])
            else:
                file_patterns.append(line)

    return (dir_patterns, file_patterns)


def _do_request(method, url, headers, data=None, query_params=None, timeout=20):
    """
    Performs an HTTP request

    :param method:
        A unicode string of 'POST' or 'PUT'

    :param url;
        A unicode string of the URL to request

    :param headers:
        A dict of unicode strings, where keys are header names and values are
        the header values.

    :param data:
        A dict of unicode strings (to be encoded as
        application/x-www-form-urlencoded), or a byte string of data.

    :param query_params:
        A dict of unicode keys and values to pass as query params

    :param timeout:
        An integer number of seconds to use as the timeout

    :return:
        A 3-element tuple:
         - 0: A unicode string of the response content-type
         - 1: A unicode string of the response encoding, or None
         - 2: A byte string of the response body
    """

    if query_params:
        url += '?' + urlencode(query_params).replace('+', '%20')

    if isinstance(data, dict):
        data_bytes = {}
        for key in data:
            data_bytes[key.encode('utf-8')] = data[key].encode('utf-8')
        data = urlencode(data_bytes)
        headers['Content-Type'] = 'application/x-www-form-urlencoded'
    if isinstance(data, str_cls):
        raise TypeError('data must be a byte string')

    try:
        tempfd, tempf_path = tempfile.mkstemp('-coverage')
        os.write(tempfd, data or b'')
        os.close(tempfd)

        if sys.platform == 'win32':
            powershell_exe = os.path.join('system32\\WindowsPowerShell\\v1.0\\powershell.exe')
            code = "[System.Net.ServicePointManager]::SecurityProtocol = [System.Net.SecurityProtocolType]::Tls12;"
            code += "$wc = New-Object Net.WebClient;"
            for key in headers:
                code += "$wc.Headers.add('%s','%s');" % (key, headers[key])
            code += "$out = $wc.UploadFile('%s', '%s', '%s');" % (url, method, tempf_path)
            code += "[System.Text.Encoding]::GetEncoding('ISO-8859-1').GetString($wc.ResponseHeaders.ToByteArray())"

            # To properly obtain bytes, we use BitConverter to get hex dash
            # encoding (e.g. AE-09-3F) and they decode in python
            code += " + [System.BitConverter]::ToString($out);"
            stdout, stderr = _execute(
                [powershell_exe, '-Command', code],
                os.getcwd(),
                re.compile(r'Unable to connect to|TLS|Internal Server Error'),
                6
            )
            if stdout[-2:] == b'\r\n' and b'\r\n\r\n' in stdout:
                # An extra trailing crlf is added at the end by powershell
                stdout = stdout[0:-2]
                parts = stdout.split(b'\r\n\r\n', 1)
                if len(parts) == 2:
                    stdout = parts[0] + b'\r\n\r\n' + codecs.decode(parts[1].replace(b'-', b''), 'hex_codec')

        else:
            args = [
                'curl',
                '--http1.1',
                '--connect-timeout', '5',
                '--request',
                method,
                '--location',
                '--silent',
                '--show-error',
                '--include',
                # Prevent curl from asking for an HTTP "100 Continue" response
                '--header', 'Expect:'
            ]
            for key in headers:
                args.append('--header')
                args.append("%s: %s" % (key, headers[key]))
            args.append('--data-binary')
            args.append('@%s' % tempf_path)
            args.append(url)
            stdout, stderr = _execute(
                args,
                os.getcwd(),
                re.compile(r'Failed to connect to|TLS|SSLRead|outstanding|cleanly|timed out'),
                6
            )
    finally:
        if tempf_path and os.path.exists(tempf_path):
            os.remove(tempf_path)

    if len(stderr) > 0:
        raise URLError("Error %sing %s:\n%s" % (method, url, stderr))

    parts = stdout.split(b'\r\n\r\n', 1)
    if len(parts) != 2:
        raise URLError("Error %sing %s, response data malformed:\n%s" % (method, url, stdout))
    header_block, body = parts

    content_type_header = None
    content_len_header = None
    for hline in header_block.decode('iso-8859-1').splitlines():
        hline_parts = hline.split(':', 1)
        if len(hline_parts) != 2:
            continue
        name, val = hline_parts
        name = name.strip().lower()
        val = val.strip()
        if name == 'content-type':
            content_type_header = val
        if name == 'content-length':
            content_len_header = val

    if content_type_header is None and content_len_header != '0':
        raise URLError("Error %sing %s, no content-type header:\n%s" % (method, url, stdout))

    if content_type_header is None:
        content_type = 'text/plain'
        encoding = 'utf-8'
    else:
        content_type, params = cgi.parse_header(content_type_header)
        encoding = params.get('charset')

    return (content_type, encoding, body)


def _execute(params, cwd, retry=None, retries=0, backoff=2):
    """
    Executes a subprocess

    :param params:
        A list of the executable and arguments to pass to it

    :param cwd:
        The working directory to execute the command in

    :param retry:
        If this string is present in stderr, or regex pattern matches stderr, retry the operation

    :param retries:
        An integer number of times to retry

    :return:
        A 2-element tuple of (stdout, stderr)
    """

    proc = subprocess.Popen(
        params,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        cwd=cwd
    )
    stdout, stderr = proc.communicate()
    code = proc.wait()
    if code != 0:
        if retry and retries > 0:
            stderr_str = stderr.decode('utf-8')
            if isinstance(retry, Pattern):
                if retry.search(stderr_str) is not None:
                    time.sleep(backoff)
                    return _execute(params, cwd, retry, retries - 1, backoff * 2)
            elif retry in stderr_str:
                time.sleep(backoff)
                return _execute(params, cwd, retry, retries - 1, backoff * 2)
        e = OSError('subprocess exit code for "%s" was %d: %s' % (' '.join(params), code, stderr))
        e.stdout = stdout
        e.stderr = stderr
        raise e
    return (stdout, stderr)


if __name__ == '__main__':
    _codecov_submit()



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import unittest
import re
import sys
import warnings

from . import requires_oscrypto
from ._import import _preload

from tests import test_classes

if sys.version_info < (3,):
    range = xrange  # noqa
    from cStringIO import StringIO
else:
    from io import StringIO


run_args = [
    {
        'name': 'regex',
        'kwarg': 'matcher',
    },
    {
        'name': 'repeat_count',
        'kwarg': 'repeat',
        'cast': 'int',
    },
]


def run(matcher=None, repeat=1, ci=False):
    """
    Runs the tests

    :param matcher:
        A unicode string containing a regular expression to use to filter test
        names by. A value of None will cause no filtering.

    :param repeat:
        An integer - the number of times to run the tests

    :param ci:
        A bool, indicating if the tests are being run as part of CI

    :return:
        A bool - if the tests succeeded
    """

    _preload(requires_oscrypto, not ci)

    warnings.filterwarnings("error")

    loader = unittest.TestLoader()
    # We have to manually track the list of applicable tests because for
    # some reason with Python 3.4 on Windows, the tests in a suite are replaced
    # with None after being executed. This breaks the repeat functionality.
    test_list = []
    for test_class in test_classes():
        if matcher:
            names = loader.getTestCaseNames(test_class)
            for name in names:
                if re.search(matcher, name):
                    test_list.append(test_class(name))
        else:
            test_list.append(loader.loadTestsFromTestCase(test_class))

    stream = sys.stdout
    verbosity = 1
    if matcher and repeat == 1:
        verbosity = 2
    elif repeat > 1:
        stream = StringIO()

    for _ in range(0, repeat):
        suite = unittest.TestSuite()
        for test in test_list:
            suite.addTest(test)
        result = unittest.TextTestRunner(stream=stream, verbosity=verbosity).run(suite)

        if len(result.errors) > 0 or len(result.failures) > 0:
            if repeat > 1:
                print(stream.getvalue())
            return False

        if repeat > 1:
            stream.truncate(0)

    return True



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import sys
import unittest

from . import test_classes


suite = unittest.TestSuite()
loader = unittest.TestLoader()
for test_class in test_classes():
    suite.addTest(loader.loadTestsFromTestCase(test_class))
unittest.TextTestRunner(stream=sys.stdout, verbosity=2).run(suite)



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os
import shutil

from . import build_root, other_packages


def run():
    """
    Cleans up CI dependencies - used for persistent GitHub Actions
    Runners since they don't clean themselves up.
    """

    print("Removing ci dependencies")
    deps_dir = os.path.join(build_root, 'modularcrypto-deps')
    if os.path.exists(deps_dir):
        shutil.rmtree(deps_dir, ignore_errors=True)

    print("Removing modularcrypto packages")
    for other_package in other_packages:
        pkg_dir = os.path.join(build_root, other_package)
        if os.path.exists(pkg_dir):
            shutil.rmtree(pkg_dir, ignore_errors=True)
    print()

    return True



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function


import socket
import select
import threading


def listen(server, on_read):
    sock, addr = server.accept()

    timeout = 10
    try:
        read_ready, _, _ = select.select([sock], [], [], timeout)
        while len(read_ready):
            data = sock.recv(8192)
            if on_read(sock, data):
                read_ready, _, _ = select.select([sock], [], [], 0)
            else:
                read_ready = []
    except (socket.error, select.error, OSError, ValueError):
        pass
    try:
        sock.shutdown(socket.SHUT_RDWR)
    except (socket.error, OSError, ValueError):
        pass
    sock.close()


def make_socket_server(port, on_read=None):
    server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind(('', port))
    server.listen(1)
    t = threading.Thread(
        target=listen,
        args=(server, on_read)
    )
    t.start()
    return server



# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

import os
import platform
import sys
import subprocess


run_args = [
    {
        'name': 'cffi',
        'kwarg': 'cffi',
    },
    {
        'name': 'openssl',
        'kwarg': 'openssl',
    },
    {
        'name': 'winlegacy',
        'kwarg': 'winlegacy',
    },
]


def _write_env(env, key, value):
    sys.stdout.write("%s: %s\n" % (key, value))
    sys.stdout.flush()
    if sys.version_info < (3,):
        env[key.encode('utf-8')] = value.encode('utf-8')
    else:
        env[key] = value


def run(**_):
    """
    Runs CI, setting various env vars

    :return:
        A bool - if the CI ran successfully
    """

    env = os.environ.copy()
    options = set(sys.argv[2:])

    newline = False
    if 'cffi' not in options:
        _write_env(env, 'OSCRYPTO_USE_CTYPES', 'true')
        newline = True
    if 'openssl' in options and sys.platform == 'darwin':
        mac_version_info = tuple(map(int, platform.mac_ver()[0].split('.')[:2]))
        if mac_version_info < (10, 15):
            _write_env(env, 'OSCRYPTO_USE_OPENSSL', '/usr/lib/libcrypto.dylib,/usr/lib/libssl.dylib')
        else:
            _write_env(env, 'OSCRYPTO_USE_OPENSSL', '/usr/lib/libcrypto.35.dylib,/usr/lib/libssl.35.dylib')
        newline = True
    if 'openssl3' in options and sys.platform == 'darwin':
        _write_env(
            env,
            'OSCRYPTO_USE_OPENSSL',
            '/usr/local/opt/openssl@3/lib/libcrypto.dylib,/usr/local/opt/openssl@3/lib/libssl.dylib'
        )
    if 'winlegacy' in options:
        _write_env(env, 'OSCRYPTO_USE_WINLEGACY', 'true')
        newline = True

    if newline:
        sys.stdout.write("\n")

    proc = subprocess.Popen(
        [
            sys.executable,
            'run.py',
            'ci',
        ],
        env=env
    )
    proc.communicate()
    return proc.returncode == 0
